#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Polars materializer."""

import os
from typing import Any, ClassVar, Tuple, Type, Union

import polars as pl
import pyarrow as pa  # type: ignore
import pyarrow.parquet as pq  # type: ignore

from zenml.enums import ArtifactType
from zenml.materializers.base_materializer import BaseMaterializer
from zenml.utils import io_utils


class PolarsMaterializer(BaseMaterializer):
    """Materializer to read/write Polars dataframes."""

    ASSOCIATED_TYPES: ClassVar[Tuple[Type[Any], ...]] = (
        pl.DataFrame,
        pl.Series,
    )
    ASSOCIATED_ARTIFACT_TYPE = ArtifactType.DATA

    def load(self, data_type: Type[Any]) -> Any:
        """Reads and returns Polars data after copying it to temporary path.

        Args:
            data_type: The type of the data to read.

        Returns:
            A Polars data frame or series.
        """
        with self.get_temporary_directory(delete_at_exit=True) as temp_dir:
            io_utils.copy_dir(self.uri, temp_dir)

            # Load the data from the temporary directory
            table = pq.read_table(
                os.path.join(temp_dir, "dataframe.parquet").replace("\\", "/")
            )

            # If the data is of type pl.Series, convert it back to a pyarrow array
            # instead of a table.
            if (
                table.schema.metadata
                and b"zenml_is_pl_series" in table.schema.metadata
            ):
                isinstance_bytes = table.schema.metadata[b"zenml_is_pl_series"]
                isinstance_series = bool.from_bytes(isinstance_bytes, "big")
                if isinstance_series:
                    table = table.column(0)

            # Convert the table to a Polars data frame or series
            data = pl.from_arrow(table)

            return data

    def save(self, data: Union[pl.DataFrame, pl.Series]) -> None:
        """Writes Polars data to the artifact store.

        Args:
            data: The data to write.

        Raises:
            TypeError: If the data is not of type pl.DataFrame or pl.Series.
        """
        # Data type check
        if not isinstance(data, self.ASSOCIATED_TYPES):
            raise TypeError(
                f"Expected data of type {self.ASSOCIATED_TYPES}, "
                f"got {type(data)}"
            )

        # Convert the data to an Apache Arrow Table
        if isinstance(data, pl.DataFrame):
            table = data.to_arrow()
        else:
            # Construct a PyArrow Table with schema from the individual pl.Series
            # array if it is a single pl.Series.
            array = data.to_arrow()
            table = pa.Table.from_arrays([array], names=[data.name])

        # Register whether data is of type pl.Series, so that the materializer read step can
        # convert it back appropriately.
        isinstance_bytes = isinstance(data, pl.Series).to_bytes(1, "big")
        table = table.replace_schema_metadata(
            {b"zenml_is_pl_series": isinstance_bytes}
        )

        with self.get_temporary_directory(delete_at_exit=True) as temp_dir:
            # Write the table to a Parquet file
            path = os.path.join(temp_dir, "dataframe.parquet").replace(
                "\\", "/"
            )
            pq.write_table(table, path)  # Uses lz4 compression by default
            io_utils.copy_dir(temp_dir, self.uri)



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Endpoint definitions for deployments."""

from uuid import UUID

from fastapi import (
    APIRouter,
    Depends,
    Security,
)

from zenml.constants import (
    API,
    DEPLOYMENTS,
    VERSION_1,
)
from zenml.models import (
    DeploymentFilter,
    DeploymentRequest,
    DeploymentResponse,
    DeploymentUpdate,
)
from zenml.models.v2.base.page import Page
from zenml.zen_server.auth import AuthContext, authorize
from zenml.zen_server.exceptions import error_response
from zenml.zen_server.rbac.endpoint_utils import (
    verify_permissions_and_create_entity,
    verify_permissions_and_delete_entity,
    verify_permissions_and_get_entity,
    verify_permissions_and_list_entities,
    verify_permissions_and_update_entity,
)
from zenml.zen_server.rbac.models import ResourceType
from zenml.zen_server.utils import (
    async_fastapi_endpoint_wrapper,
    make_dependable,
    zen_store,
)

router = APIRouter(
    prefix=API + VERSION_1 + DEPLOYMENTS,
    tags=["deployments"],
    responses={401: error_response, 403: error_response},
)


@router.post(
    "",
    responses={401: error_response, 409: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def create_deployment(
    deployment: DeploymentRequest,
    _: AuthContext = Security(authorize),
) -> DeploymentResponse:
    """Creates a deployment.

    Args:
        deployment: Deployment to create.

    Returns:
        The created deployment.
    """
    return verify_permissions_and_create_entity(
        request_model=deployment,
        create_method=zen_store().create_deployment,
    )


@router.get(
    "",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def list_deployments(
    deployment_filter_model: DeploymentFilter = Depends(
        make_dependable(DeploymentFilter)
    ),
    hydrate: bool = False,
    _: AuthContext = Security(authorize),
) -> Page[DeploymentResponse]:
    """Gets a list of deployments.

    Args:
        deployment_filter_model: Filter model used for pagination, sorting,
            filtering.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        List of deployment objects matching the filter criteria.
    """
    return verify_permissions_and_list_entities(
        filter_model=deployment_filter_model,
        resource_type=ResourceType.DEPLOYMENT,
        list_method=zen_store().list_deployments,
        hydrate=hydrate,
    )


@router.get(
    "/{deployment_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def get_deployment(
    deployment_id: UUID,
    hydrate: bool = True,
    _: AuthContext = Security(authorize),
) -> DeploymentResponse:
    """Gets a specific deployment using its unique id.

    Args:
        deployment_id: ID of the deployment to get.
        hydrate: Flag deciding whether to hydrate the output model(s)
            by including metadata fields in the response.

    Returns:
        A specific deployment object.
    """
    return verify_permissions_and_get_entity(
        id=deployment_id,
        get_method=zen_store().get_deployment,
        hydrate=hydrate,
    )


@router.put(
    "/{deployment_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper(deduplicate=True)
def update_deployment(
    deployment_id: UUID,
    deployment_update: DeploymentUpdate,
    _: AuthContext = Security(authorize),
) -> DeploymentResponse:
    """Updates a specific deployment.

    Args:
        deployment_id: ID of the deployment to update.
        deployment_update: Update model for the deployment.

    Returns:
        The updated deployment.
    """
    return verify_permissions_and_update_entity(
        id=deployment_id,
        update_model=deployment_update,
        get_method=zen_store().get_deployment,
        update_method=zen_store().update_deployment,
    )


@router.delete(
    "/{deployment_id}",
    responses={401: error_response, 404: error_response, 422: error_response},
)
@async_fastapi_endpoint_wrapper
def delete_deployment(
    deployment_id: UUID,
    _: AuthContext = Security(authorize),
) -> None:
    """Deletes a specific deployment.

    Args:
        deployment_id: ID of the deployment to delete.
    """
    verify_permissions_and_delete_entity(
        id=deployment_id,
        get_method=zen_store().get_deployment,
        delete_method=zen_store().delete_deployment,
    )



"""Release [0.93.3].

Revision ID: 0.93.3
Revises: 8e36b8227160
Create Date: 2026-02-18 13:41:36.511218

"""

# revision identifiers, used by Alembic.
revision = "0.93.3"
down_revision = "8e36b8227160"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    pass


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    pass



"""Add run_once_start_time to Schedule [1d74e596abb8].

Revision ID: 1d74e596abb8
Revises: 0.56.2
Create Date: 2024-04-02 11:35:28.755331

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "1d74e596abb8"
down_revision = "0.56.2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    """Upgrade database schema and/or data, creating a new revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("run_once_start_time", sa.DateTime(), nullable=True)
        )

    # ### end Alembic commands ###


def downgrade() -> None:
    """Downgrade database schema and/or data back to the previous revision."""
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table("schedule", schema=None) as batch_op:
        batch_op.drop_column("run_once_start_time")

    # ### end Alembic commands ###



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes orchestrator flavor."""

from typing import TYPE_CHECKING, Any, Dict, List, Optional, Type

from pydantic import (
    Field,
    NonNegativeInt,
    PositiveFloat,
    PositiveInt,
    field_validator,
)

from zenml.config.base_settings import BaseSettings
from zenml.constants import KUBERNETES_CLUSTER_RESOURCE_TYPE
from zenml.integrations.kubernetes import KUBERNETES_ORCHESTRATOR_FLAVOR
from zenml.integrations.kubernetes.pod_settings import KubernetesPodSettings
from zenml.logger import get_logger
from zenml.models import ServiceConnectorRequirements
from zenml.orchestrators import BaseOrchestratorConfig, BaseOrchestratorFlavor
from zenml.utils import deprecation_utils

if TYPE_CHECKING:
    from zenml.integrations.kubernetes.orchestrators import (
        KubernetesOrchestrator,
    )

logger = get_logger(__name__)


class KubernetesOrchestratorSettings(BaseSettings):
    """Settings for the Kubernetes orchestrator.

    Configuration options for how pipelines are executed on Kubernetes clusters.
    Field descriptions are defined inline using Field() descriptors.
    """

    synchronous: bool = Field(
        default=True,
        description="Whether to wait for all pipeline steps to complete. "
        "When `False`, the client returns immediately and execution continues asynchronously.",
    )
    service_account_name: Optional[str] = Field(
        default=None,
        description="Kubernetes service account for the orchestrator pod. "
        "If not specified, creates a new account with 'edit' permissions.",
    )
    step_pod_service_account_name: Optional[str] = Field(
        default=None,
        description="Kubernetes service account for step execution pods. "
        "Uses the default service account if not specified.",
    )
    privileged: bool = Field(
        default=False,
        description="Whether to run containers in privileged mode with extended permissions.",
    )
    pod_settings: Optional[KubernetesPodSettings] = Field(
        default=None,
        description="Pod configuration for step execution containers.",
    )
    orchestrator_pod_settings: Optional[KubernetesPodSettings] = Field(
        default=None,
        description="Pod configuration for the orchestrator container that launches step pods.",
    )
    job_name_prefix: Optional[str] = Field(
        default=None,
        description="Prefix for the job name.",
    )
    max_parallelism: Optional[PositiveInt] = Field(
        default=None,
        description="Maximum number of step pods to run concurrently. No limit if not specified.",
    )
    successful_jobs_history_limit: Optional[NonNegativeInt] = Field(
        default=None,
        description="Number of successful scheduled jobs to retain in history.",
    )
    failed_jobs_history_limit: Optional[NonNegativeInt] = Field(
        default=None,
        description="Number of failed scheduled jobs to retain in history.",
    )
    concurrency_policy: Optional[str] = Field(
        default=None,
        description="CronJob concurrency policy for scheduled pipelines. "
        "Controls whether concurrent job executions are allowed. "
        "Valid values: 'Allow', 'Forbid', 'Replace'. "
        "Only applies when a pipeline has a cron schedule",
    )
    starting_deadline_seconds: Optional[NonNegativeInt] = Field(
        default=None,
        description="CronJob starting deadline in seconds for scheduled "
        "pipelines. If a scheduled run misses its trigger time, it can "
        "still start within this window. Only applies when a pipeline "
        "has a cron schedule",
    )
    ttl_seconds_after_finished: Optional[NonNegativeInt] = Field(
        default=None,
        description="Seconds to keep finished jobs before automatic cleanup.",
    )
    active_deadline_seconds: Optional[NonNegativeInt] = Field(
        default=None,
        description="Job deadline in seconds. If the job doesn't finish "
        "within this time, it will be terminated.",
    )
    backoff_limit_margin: NonNegativeInt = Field(
        default=0,
        description="The value to add to the backoff limit in addition "
        "to the step retries. The retry configuration defined on the step "
        "defines the maximum number of retries that the server will accept "
        "for a step. For this orchestrator, this controls how often the "
        "job running the step will try to start the step pod. There are some "
        "circumstances however where the job will start the pod, but the pod "
        "doesn't actually get to the point of running the step. That means "
        "the server will not receive the maximum amount of retry requests, "
        "which in turn causes other inconsistencies like wrong step statuses. "
        "To mitigate this, this attribute allows to add a margin to the "
        "backoff limit. This means that the job will retry the pod startup "
        "for the configured amount of times plus the margin, which increases "
        "the chance of the server receiving the maximum amount of retry "
        "requests.",
    )
    orchestrator_job_backoff_limit: NonNegativeInt = Field(
        default=3,
        description="The backoff limit for the orchestrator job.",
    )
    fail_on_container_waiting_reasons: Optional[List[str]] = Field(
        default=[
            "InvalidImageName",
            "ErrImagePull",
            "ImagePullBackOff",
            "CreateContainerConfigError",
        ],
        description="List of container waiting reasons that should cause the "
        "job to fail immediately. This should be set to a list of "
        "nonrecoverable reasons, which if found in any "
        "`pod.status.containerStatuses[*].state.waiting.reason` of a job pod, "
        "should cause the job to fail immediately.",
    )
    job_monitoring_interval: PositiveFloat = Field(
        default=3,
        description="The interval in seconds to monitor the job. Each interval "
        "is used to check for container issues for the job pods.",
    )
    job_monitoring_delay: PositiveFloat = Field(
        default=0.0,
        description="The delay in seconds to wait between monitoring active "
        "step jobs. This can be used to reduce load on the Kubernetes API "
        "server.",
    )
    interrupt_check_interval: PositiveFloat = Field(
        default=1.0,
        description="The interval in seconds to check for run interruptions.",
        ge=0.5,
    )
    pod_failure_policy: Optional[Dict[str, Any]] = Field(
        default=None,
        description="The pod failure policy to use for the job that is "
        "executing the step.",
    )
    prevent_orchestrator_pod_caching: bool = Field(
        default=False,
        description="Whether to disable caching optimization in the orchestrator pod.",
    )
    always_build_pipeline_image: bool = Field(
        default=False,
        description="If `True`, the orchestrator will always build the pipeline image, "
        "even if all steps have a custom build.",
    )
    pod_stop_grace_period: PositiveInt = Field(
        default=30,
        description="When stopping a pipeline run, the amount of seconds to wait for a step pod to shutdown gracefully.",
    )
    api_request_timeout: Optional[PositiveInt] = Field(
        default=None,
        description="Timeout for API requests in seconds. If not specified, no explicit timeout will be set. ",
    )

    # Deprecated fields
    timeout: Optional[int] = Field(
        default=None,
        deprecated=True,
        description="DEPRECATED/UNUSED.",
    )
    stream_step_logs: Optional[bool] = Field(
        default=None,
        deprecated=True,
        description="DEPRECATED/UNUSED.",
    )
    pod_startup_timeout: Optional[int] = Field(
        default=None,
        description="DEPRECATED/UNUSED.",
        deprecated=True,
    )
    pod_failure_max_retries: Optional[int] = Field(
        default=None,
        description="DEPRECATED/UNUSED.",
        deprecated=True,
    )
    pod_failure_retry_delay: Optional[int] = Field(
        default=None,
        description="DEPRECATED/UNUSED.",
        deprecated=True,
    )
    pod_failure_backoff: Optional[float] = Field(
        default=None,
        description="DEPRECATED/UNUSED.",
        deprecated=True,
    )
    pod_name_prefix: Optional[str] = Field(
        default=None,
        deprecated=True,
        description="DEPRECATED/UNUSED.",
    )

    _deprecation_validator = deprecation_utils.deprecate_pydantic_attributes(
        "timeout",
        "stream_step_logs",
        "pod_startup_timeout",
        "pod_failure_max_retries",
        "pod_failure_retry_delay",
        "pod_failure_backoff",
        ("pod_name_prefix", "job_name_prefix"),
    )

    @field_validator("concurrency_policy", mode="before")
    @classmethod
    def _validate_concurrency_policy(cls, value: Any) -> Any:
        """Validates and normalizes the CronJob concurrency policy.

        Args:
            value: The concurrency policy value.

        Returns:
            The normalized value.

        Raises:
            ValueError: If the value is not a valid concurrency policy.
        """
        if value is None:
            return value

        allowed = {"Allow", "Forbid", "Replace"}
        # Normalize: capitalize first letter, lowercase rest
        normalized = str(value).strip().capitalize()
        if normalized not in allowed:
            raise ValueError(
                f"Invalid concurrency policy '{value}'. "
                f"Must be one of: {', '.join(sorted(allowed))}"
            )
        return normalized

    @field_validator("pod_failure_policy", mode="before")
    @classmethod
    def _convert_pod_failure_policy(cls, value: Any) -> Any:
        """Converts Kubernetes pod failure policy to a dict.

        Args:
            value: The pod failure policy value.

        Returns:
            The converted value.
        """
        from kubernetes.client.models import V1PodFailurePolicy

        from zenml.integrations.kubernetes import serialization_utils

        if isinstance(value, V1PodFailurePolicy):
            return serialization_utils.serialize_kubernetes_model(value)
        else:
            return value

    @field_validator("api_request_timeout", mode="before")
    @classmethod
    def api_request_timeout_validator(cls, value: Any) -> Any:
        """Validates API request timeout.

        Args:
            value: The API request timeout value.

        Returns:
            The validated value in integer format.
        """
        if isinstance(value, float):
            logger.warning(
                f"Converted `api_request_timeout` from float to int: {value} -> {int(value)}. "
                "Consider updating the Orchestrator settings."
            )
            return int(value)
        return value


class KubernetesOrchestratorConfig(
    BaseOrchestratorConfig, KubernetesOrchestratorSettings
):
    """Configuration for the Kubernetes orchestrator."""

    incluster: bool = Field(
        False,
        description="If `True`, the orchestrator will attempt to load the "
        "in-cluster Kubernetes configuration and run the pipeline inside the "
        "same cluster in which it itself is running. This requires the client "
        "to run in a Kubernetes pod itself. If this fails, the orchestrator "
        "will fall back to using the linked service connector or the "
        "configured `kubernetes_context` configuration if provided, in that "
        "order.",
    )
    kubernetes_context: Optional[str] = Field(
        None,
        description="Name of a Kubernetes context to run pipelines in. "
        "If the stack component is linked to a Kubernetes service connector, "
        "this field is ignored. Otherwise, it is mandatory.",
    )
    kubernetes_namespace: str = Field(
        "zenml",
        description="Name of the Kubernetes namespace to be used. "
        "If not provided, `zenml` namespace will be used.",
    )
    local: bool = Field(
        False,
        description="If `True`, the orchestrator will assume it is connected to a "
        "local kubernetes cluster and will perform additional validations and "
        "operations to allow using the orchestrator in combination with other "
        "local stack components that store data in the local filesystem "
        "(i.e. it will mount the local stores directory into the pipeline containers).",
    )
    skip_local_validations: bool = Field(
        False, description="If `True`, the local validations will be skipped."
    )
    parallel_step_startup_waiting_period: Optional[float] = Field(
        None,
        description="How long to wait in between starting parallel steps. "
        "This can be used to distribute server load when running pipelines "
        "with a huge amount of parallel steps.",
    )
    pass_zenml_token_as_secret: bool = Field(
        False,
        description="If `True`, the ZenML token will be passed as a Kubernetes secret "
        "to the pods. For this to work, the Kubernetes client must have permissions "
        "to create secrets in the namespace.",
    )
    skip_owner_references: bool = Field(
        False,
        description="If `True`, the orchestrator will not alter the owner "
        "references on the step jobs.",
    )

    @property
    def is_remote(self) -> bool:
        """Checks if this stack component is running remotely.

        This designation is used to determine if the stack component can be
        used with a local ZenML database or if it requires a remote ZenML
        server.

        Returns:
            True if this config is for a remote component, False otherwise.
        """
        return not self.local

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return self.local

    @property
    def is_synchronous(self) -> bool:
        """Whether the orchestrator runs synchronous or not.

        Returns:
            Whether the orchestrator runs synchronous or not.
        """
        return self.synchronous

    @property
    def is_schedulable(self) -> bool:
        """Whether the orchestrator is schedulable or not.

        Returns:
            Whether the orchestrator is schedulable or not.
        """
        return True

    @property
    def supports_client_side_caching(self) -> bool:
        """Whether the orchestrator supports client side caching.

        Returns:
            Whether the orchestrator supports client side caching.
        """
        # The Kubernetes orchestrator starts step pods from a pipeline pod.
        # This is currently not supported when using client-side caching.
        return False

    @property
    def handles_step_retries(self) -> bool:
        """Whether the orchestrator handles step retries.

        Returns:
            Whether the orchestrator handles step retries.
        """
        return True


class KubernetesOrchestratorFlavor(BaseOrchestratorFlavor):
    """Kubernetes orchestrator flavor."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return KUBERNETES_ORCHESTRATOR_FLAVOR

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            resource_type=KUBERNETES_CLUSTER_RESOURCE_TYPE,
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/kubernetes.png"

    @property
    def config_class(self) -> Type[KubernetesOrchestratorConfig]:
        """Returns `KubernetesOrchestratorConfig` config class.

        Returns:
                The config class.
        """
        return KubernetesOrchestratorConfig

    @property
    def implementation_class(self) -> Type["KubernetesOrchestrator"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.kubernetes.orchestrators import (
            KubernetesOrchestrator,
        )

        return KubernetesOrchestrator



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of utils specific to the MLflow integration."""

import mlflow
from mlflow.entities import Run
from packaging import version

from zenml.client import Client
from zenml.logger import get_logger

logger = get_logger(__name__)

ZENML_TAG_KEY = "zenml"


def is_mlflow_3x() -> bool:
    """Check if MLflow version is 3.x or higher.

    Returns:
        True if MLflow version is 3.x or higher, False otherwise.
    """
    return version.parse(mlflow.version.VERSION) >= version.parse("3.0.0")


def get_model_uri_from_version(
    mlflow_model_version: "mlflow.entities.model_registry.ModelVersion",
) -> str:
    """Get the model URI from an MLflow model version.

    This function handles the differences between MLflow 2.x and 3.x:
    - In MLflow 3.x, models are stored separately and we construct a runs:/ URI
    - In MLflow 2.x, we use the source path directly

    Args:
        mlflow_model_version: The MLflow model version object.

    Returns:
        The model URI that can be used to load the model.
    """
    if is_mlflow_3x() and mlflow_model_version.run_id:
        source_path = mlflow_model_version.source
        if source_path and "/artifacts/" in source_path:
            artifact_name = source_path.split("/artifacts/")[-1]
            return f"runs:/{mlflow_model_version.run_id}/{artifact_name}"
        return source_path or ""
    return mlflow_model_version.source or ""


def build_runs_uri(run_id: str, artifact_path: str) -> str:
    """Build a runs:/ URI from a run ID and artifact path.

    Args:
        run_id: The MLflow run ID.
        artifact_path: The artifact path (e.g., "model").

    Returns:
        The constructed URI.
    """
    return f"runs:/{run_id}/{artifact_path}"


def get_missing_mlflow_experiment_tracker_error() -> ValueError:
    """Returns description of how to add an MLflow experiment tracker to your stack.

    Returns:
        ValueError: If no MLflow experiment tracker is registered in the active stack.
    """
    return ValueError(
        "The active stack needs to have a MLflow experiment tracker "
        "component registered to be able to track experiments using "
        "MLflow. You can create a new stack with a MLflow experiment "
        "tracker component or update your existing stack to add this "
        "component, e.g.:\n\n"
        "  'zenml experiment-tracker register mlflow_tracker "
        "--type=mlflow'\n"
        "  'zenml stack register stack-name -e mlflow_tracker ...'\n"
    )


def get_tracking_uri() -> str:
    """Gets the MLflow tracking URI from the active experiment tracking stack component.

    # noqa: DAR401

    Returns:
        MLflow tracking URI.
    """
    from zenml.integrations.mlflow.experiment_trackers.mlflow_experiment_tracker import (
        MLFlowExperimentTracker,
    )

    tracker = Client().active_stack.experiment_tracker
    if tracker is None or not isinstance(tracker, MLFlowExperimentTracker):
        raise get_missing_mlflow_experiment_tracker_error()

    return tracker.get_tracking_uri()


def is_zenml_run(run: Run) -> bool:
    """Checks if a MLflow run is a ZenML run or not.

    Args:
        run: The run to check.

    Returns:
        If the run is a ZenML run.
    """
    return ZENML_TAG_KEY in run.data.tags


def stop_zenml_mlflow_runs(status: str) -> None:
    """Stops active ZenML Mlflow runs.

    This function stops all MLflow active runs until no active run exists or
    a non-ZenML run is active.

    Args:
        status: The status to set the run to.
    """
    active_run = mlflow.active_run()
    while active_run:
        if is_zenml_run(active_run):
            logger.debug("Stopping mlflow run %s.", active_run.info.run_id)
            mlflow.end_run(status=status)
            active_run = mlflow.active_run()
        else:
            break



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""MLflow model deployer flavor."""

from typing import TYPE_CHECKING, Optional, Type

from pydantic import Field

from zenml.integrations.mlflow import MLFLOW_MODEL_DEPLOYER_FLAVOR
from zenml.model_deployers.base_model_deployer import (
    BaseModelDeployerConfig,
    BaseModelDeployerFlavor,
)

if TYPE_CHECKING:
    from zenml.integrations.mlflow.model_deployers import MLFlowModelDeployer


class MLFlowModelDeployerConfig(BaseModelDeployerConfig):
    """Configuration for the MLflow model deployer.

    Configuration for local MLflow model serving.
    Field descriptions are defined inline using Field() descriptors.
    """

    service_path: str = Field(
        default="",
        description="Local directory for MLflow deployment service files "
        "(configuration, PID, and logs). Uses temp directory if empty.",
    )

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return True


class MLFlowModelDeployerFlavor(BaseModelDeployerFlavor):
    """Model deployer flavor for MLflow models."""

    @property
    def name(self) -> str:
        """Name of the flavor.

        Returns:
            The name of the flavor.
        """
        return MLFLOW_MODEL_DEPLOYER_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the flavor.

        Returns:
            The display name of the flavor.
        """
        return "MLflow"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/model_deployer/mlflow.png"

    @property
    def config_class(self) -> Type[MLFlowModelDeployerConfig]:
        """Returns `MLFlowModelDeployerConfig` config class.

        Returns:
                The config class.
        """
        return MLFlowModelDeployerConfig

    @property
    def implementation_class(self) -> Type["MLFlowModelDeployer"]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        from zenml.integrations.mlflow.model_deployers import (
            MLFlowModelDeployer,
        )

        return MLFlowModelDeployer



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML Local deployer."""

import ipaddress
import os
import shutil
import subprocess
import sys
import time
from typing import (
    TYPE_CHECKING,
    Dict,
    Generator,
    List,
    Optional,
    Tuple,
    Type,
    cast,
)
from uuid import UUID

import psutil
from pydantic import BaseModel

from zenml.config.base_settings import BaseSettings
from zenml.deployers.base_deployer import (
    BaseDeployer,
    BaseDeployerConfig,
    BaseDeployerFlavor,
    BaseDeployerSettings,
)
from zenml.deployers.exceptions import (
    DeployerError,
    DeploymentDeprovisionError,
    DeploymentLogsNotFoundError,
    DeploymentNotFoundError,
    DeploymentProvisionError,
)
from zenml.deployers.server.app import (
    start_deployment_app,
)
from zenml.enums import DeploymentStatus
from zenml.logger import get_logger
from zenml.models import DeploymentOperationalState, DeploymentResponse
from zenml.utils.daemon import (
    stop_process,
)
from zenml.utils.io_utils import get_global_config_directory
from zenml.utils.networking_utils import (
    lookup_preferred_or_free_port,
)

if TYPE_CHECKING:
    from zenml.stack import Stack

logger = get_logger(__name__)

DEFAULT_TAIL_FOLLOW_LINES = 100
DEFAULT_DAEMON_STARTUP_TIMEOUT = 10


class LocalDeploymentMetadata(BaseModel):
    """Metadata for a local daemon deployment.

    Attributes:
        pid: PID of the daemon process.
        port: TCP port the app listens on.
        address: IP address the app binds to.
        log_file: Path to log file.
    """

    pid: Optional[int] = None
    port: Optional[int] = None
    address: Optional[str] = None
    log_file: Optional[str] = None

    @classmethod
    def from_deployment(
        cls, deployment: DeploymentResponse
    ) -> "LocalDeploymentMetadata":
        """Build metadata object from a deployment record.

        Args:
            deployment: The deployment to read metadata from.

        Returns:
            Parsed local deployment metadata.
        """
        return cls.model_validate(deployment.deployment_metadata or {})


class LocalDeployerSettings(BaseDeployerSettings):
    """Local deployer settings.

    Attributes:
        port: Preferred port to run on.
        allocate_port_if_busy: Whether to allocate a free port if busy.
        port_range: Range to scan when allocating a free port.
        address: Address to bind the server to.
        blocking: Whether to run the deployment in the current process instead
            of running it as a daemon process.
        auto_reload: Whether to automatically reload the deployment when the
            code changes.
    """

    port: Optional[int] = None
    allocate_port_if_busy: bool = True
    port_range: Tuple[int, int] = (8000, 65535)
    address: str = "127.0.0.1"
    blocking: bool = False
    auto_reload: bool = False


class LocalDeployerConfig(BaseDeployerConfig, LocalDeployerSettings):
    """Local deployer config."""

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component.
        """
        return True


class LocalDeployer(BaseDeployer):
    """Deployer that runs deployments as local daemon processes."""

    @property
    def settings_class(self) -> Optional[Type[BaseSettings]]:
        """Settings class for the local deployer.

        Returns:
            The settings class.
        """
        return LocalDeployerSettings

    @property
    def config(self) -> LocalDeployerConfig:
        """Returns the `LocalDeployerConfig` config.

        Returns:
            The configuration.
        """
        return cast(LocalDeployerConfig, self._config)

    # ---------- Helpers ----------
    def _runtime_dir(self, deployment_id: UUID) -> str:
        """Compute runtime directory for a deployment.

        Args:
            deployment_id: The deployment UUID.

        Returns:
            Absolute runtime directory path.
        """
        return os.path.join(
            get_global_config_directory(),
            "deployments",
            str(deployment_id),
        )

    def _pid_file_path(self, deployment_id: UUID) -> str:
        """Compute PID file path for a deployment.

        Args:
            deployment_id: The deployment UUID.

        Returns:
            Absolute PID file path.
        """
        return os.path.join(self._runtime_dir(deployment_id), "daemon.pid")

    def _log_file_path(self, deployment_id: UUID) -> str:
        """Compute log file path for a deployment.

        Args:
            deployment_id: The deployment UUID.

        Returns:
            Absolute log file path.
        """
        return os.path.join(self._runtime_dir(deployment_id), "daemon.log")

    # ---------- LCM Operations ----------
    def do_provision_deployment(
        self,
        deployment: DeploymentResponse,
        stack: "Stack",
        environment: Dict[str, str],
        secrets: Dict[str, str],
        timeout: int,
    ) -> DeploymentOperationalState:
        """Provision a local daemon deployment.

        Args:
            deployment: The deployment to run.
            stack: The active stack (unused by local deployer).
            environment: Environment variables for the app.
            secrets: Secret environment variables for the app.
            timeout: Unused for immediate daemonization.

        Returns:
            Operational state of the provisioned deployment.

        Raises:
            DeploymentProvisionError: If the daemon cannot be started.
        """
        assert deployment.snapshot, "Pipeline snapshot not found"

        child_env: Dict[str, str] = dict(os.environ)
        child_env.update(environment)
        child_env.update(secrets)

        settings = cast(
            LocalDeployerSettings,
            self.get_settings(deployment.snapshot),
        )

        existing_meta = LocalDeploymentMetadata.from_deployment(deployment)

        if existing_meta.pid:
            try:
                stop_process(existing_meta.pid)
            except Exception as e:
                logger.warning(
                    f"Failed to stop existing daemon process for deployment "
                    f"'{deployment.name}' with PID {existing_meta.pid}: {e}"
                )

        preferred_ports: List[int] = []
        if settings.port:
            preferred_ports.append(settings.port)
        if existing_meta.port:
            preferred_ports.append(existing_meta.port)

        try:
            port = lookup_preferred_or_free_port(
                preferred_ports=preferred_ports,
                allocate_port_if_busy=settings.allocate_port_if_busy,
                range=settings.port_range,
                address=settings.address,
            )
        except IOError as e:
            raise DeploymentProvisionError(str(e))

        address = settings.address
        # Validate that the address is a valid IP address
        try:
            ipaddress.ip_address(address)
        except ValueError:
            raise DeploymentProvisionError(
                f"Invalid address: {address}. Must be a valid IP address."
            )

        if address == "0.0.0.0":  # nosec
            address = "localhost"
        url = f"http://{address}:{port}"

        log_file = existing_meta.log_file or self._log_file_path(deployment.id)

        runtime_dir = self._runtime_dir(deployment.id)
        if not os.path.exists(runtime_dir):
            os.makedirs(runtime_dir, exist_ok=True)

        if settings.blocking:
            self._update_deployment(
                deployment,
                DeploymentOperationalState(
                    status=DeploymentStatus.RUNNING,
                    url=url,
                    metadata=LocalDeploymentMetadata(
                        pid=os.getpid(),
                        port=port,
                        address=settings.address,
                    ).model_dump(exclude_none=True),
                ),
            )
            start_deployment_app(
                deployment_id=deployment.id,
                host=settings.address,
                port=port,
                reload=settings.auto_reload,
            )
            self._update_deployment(
                deployment,
                DeploymentOperationalState(
                    status=DeploymentStatus.ABSENT,
                    metadata=None,
                ),
            )
            # Exiting early here because the deployment takes over the current
            # process and anything else is irrelevant.
            sys.exit(0)

        # Launch the deployment app as a background subprocess.
        python_exe = sys.executable
        module = "zenml.deployers.server.app"
        cmd = [
            python_exe,
            "-m",
            module,
            "--deployment_id",
            str(deployment.id),
            "--log_file",
            os.path.abspath(log_file),
            "--host",
            settings.address,
            "--port",
            str(port),
        ]

        if settings.auto_reload:
            cmd.append("--reload")

        try:
            os.makedirs(os.path.dirname(log_file), exist_ok=True)
            proc = subprocess.Popen(
                cmd,
                cwd=os.getcwd(),
                env=child_env,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL,
                start_new_session=True,
                close_fds=True,
            )
        except Exception as e:
            raise DeploymentProvisionError(
                f"Failed to start subprocess for deployment "
                f"'{deployment.name}': {e}"
            ) from e

        metadata = LocalDeploymentMetadata(
            pid=proc.pid,
            port=port,
            address=settings.address,
            log_file=log_file,
        )

        state = DeploymentOperationalState(
            status=DeploymentStatus.PENDING,
            url=url,
            metadata=metadata.model_dump(exclude_none=True),
        )

        return state

    def do_get_deployment_state(
        self, deployment: DeploymentResponse
    ) -> DeploymentOperationalState:
        """Get information about a local daemon deployment.

        Args:
            deployment: The deployment to inspect.

        Returns:
            Operational state of the deployment.
        """
        assert deployment.snapshot, "Pipeline snapshot not found"
        meta = LocalDeploymentMetadata.from_deployment(deployment)

        state = DeploymentOperationalState(
            status=DeploymentStatus.ERROR,
            metadata=meta.model_dump(exclude_none=True),
        )

        if not meta.pid:
            state.status = DeploymentStatus.ABSENT
            return state

        if not psutil.pid_exists(meta.pid):
            return state

        if not meta.port or not meta.address:
            return state

        # Use pending until we can confirm the daemon is reachable
        state.status = DeploymentStatus.RUNNING
        address = meta.address
        if address == "0.0.0.0":  # nosec
            address = "localhost"
        state.url = f"http://{address}:{meta.port}"

        state.metadata = meta.model_dump(exclude_none=True)

        return state

    def do_get_deployment_state_logs(
        self,
        deployment: DeploymentResponse,
        follow: bool = False,
        tail: Optional[int] = None,
    ) -> Generator[str, bool, None]:
        """Read logs from the local daemon log file.

        Args:
            deployment: The deployment to read logs for.
            follow: Stream logs if True.
            tail: Return only last N lines if set.

        Yields:
            Log lines.

        Raises:
            DeploymentLogsNotFoundError: If the log file is missing.
            DeployerError: For unexpected errors.
        """
        meta = LocalDeploymentMetadata.from_deployment(deployment)
        log_file = meta.log_file
        if not log_file or not os.path.exists(log_file):
            raise DeploymentLogsNotFoundError(
                f"Log file not found for deployment '{deployment.name}'"
            )

        try:

            def _read_tail(path: str, n: int) -> Generator[str, bool, None]:
                with open(path, "r", encoding="utf-8", errors="ignore") as f:
                    lines = f.readlines()
                    for line in lines[-n:]:
                        yield line.rstrip("\n")

            if not follow:
                if tail and tail > 0:
                    yield from _read_tail(log_file, tail)
                else:
                    with open(
                        log_file, "r", encoding="utf-8", errors="ignore"
                    ) as f:
                        for line in f:
                            yield line.rstrip("\n")
                return

            with open(log_file, "r", encoding="utf-8", errors="ignore") as f:
                if not tail:
                    tail = DEFAULT_TAIL_FOLLOW_LINES
                lines = f.readlines()
                for line in lines[-tail:]:
                    yield line.rstrip("\n")

                while True:
                    where = f.tell()
                    line = f.readline()
                    if not line:
                        time.sleep(0.2)
                        f.seek(where)
                        continue
                    yield line.rstrip("\n")

        except DeploymentLogsNotFoundError:
            raise
        except Exception as e:
            raise DeployerError(
                f"Unexpected error while reading logs for deployment "
                f"'{deployment.name}': {e}"
            ) from e

    def do_deprovision_deployment(
        self, deployment: DeploymentResponse, timeout: int
    ) -> Optional[DeploymentOperationalState]:
        """Deprovision a local daemon deployment.

        Args:
            deployment: The deployment to stop.
            timeout: Unused for local daemon stop.

        Returns:
            None, indicating immediate deletion completed.

        Raises:
            DeploymentNotFoundError: If the daemon is not found.
            DeploymentDeprovisionError: If stopping fails.
        """
        meta = LocalDeploymentMetadata.from_deployment(deployment)
        if not meta.pid:
            raise DeploymentNotFoundError(
                f"Daemon for deployment '{deployment.name}' missing."
            )

        try:
            stop_process(meta.pid)
        except Exception as e:
            raise DeploymentDeprovisionError(
                f"Failed to stop daemon for deployment '{deployment.name}': "
                f"{e}"
            ) from e
        else:
            shutil.rmtree(self._runtime_dir(deployment.id))

        return None


class LocalDeployerFlavor(BaseDeployerFlavor):
    """Flavor for the Local daemon deployer."""

    @property
    def name(self) -> str:
        """Name of the deployer flavor.

        Returns:
            Flavor name.
        """
        return "local"

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/deployer/local.png"

    @property
    def config_class(self) -> Type[BaseDeployerConfig]:
        """Config class for the flavor.

        Returns:
            The config class.
        """
        return LocalDeployerConfig

    @property
    def implementation_class(self) -> Type[LocalDeployer]:
        """Implementation class for this flavor.

        Returns:
            The implementation class.
        """
        return LocalDeployer



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Base deployment app runner."""

import os
import re
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)
from uuid import UUID

from asgiref.compatibility import guarantee_single_callable
from asgiref.typing import (
    ASGIApplication,
    ASGIReceiveCallable,
    ASGISendCallable,
    ASGISendEvent,
    Scope,
)

from zenml.client import Client
from zenml.config import (
    AppExtensionSpec,
    DeploymentDefaultEndpoints,
    DeploymentDefaultMiddleware,
    DeploymentSettings,
    EndpointMethod,
    EndpointSpec,
    MiddlewareSpec,
)
from zenml.config.source import SourceOrObject
from zenml.deployers.server.adapters import (
    EndpointAdapter,
    MiddlewareAdapter,
)
from zenml.deployers.server.extensions import BaseAppExtension
from zenml.deployers.server.models import (
    BaseDeploymentInvocationRequest,
    BaseDeploymentInvocationResponse,
)
from zenml.deployers.server.service import (
    BasePipelineDeploymentService,
    PipelineDeploymentService,
)
from zenml.integrations.registry import integration_registry
from zenml.logger import get_logger
from zenml.models.v2.core.deployment import DeploymentResponse
from zenml.utils import source_utils
from zenml.utils.daemon import setup_daemon

if TYPE_CHECKING:
    from secure import Secure

logger = get_logger(__name__)

ZENML_DEPLOYMENT_ID_ENV_VAR = "ZENML_DEPLOYMENT_ID"


class BaseDeploymentAppRunner(ABC):
    """Base class for deployment app runners.

    This class is responsible for building and running the ASGI compatible web
    application (e.g. FastAPI, Django, Flask, Falcon, Quart, BlackSheep, etc.) and the
    associated deployment service for the pipeline deployment. It also acts as
    a adaptation layer between the REST API interface and deployment service to
    preserve the following separation of concerns between the two components:

    * the ASGI application is responsible for handling the HTTP requests and
    responses to the user
    * the deployment service is responsible for handling the business logic

    The deployment service code should be free of any ASGI application specific
    code and concerns and vice-versa. This allows them to be independently
    extendable and easily swappable.

    Implementations of this class must use the deployment and its settings to
    configure and run the web application (e.g. FastAPI, Flask, Falcon, Quart,
    BlackSheep, etc.) that wraps the deployment service according to the user's
    specifications, particularly concerning the following:

    * exposed endpoints (URL paths, methods, input/output models)
    * middleware (CORS, authentication, logging, etc.)
    * error handling
    * lifecycle management (startup, shutdown)
    * custom hooks (startup, shutdown)
    * app configuration (workers, host, port, thread pool size, etc.)

    The following methods must be provided by implementations of this class:

    * flavor: Return the flavor class associated with this deployment
    application runner.
    * build: Build and return an ASGI compatible web application (i.e. an
    ASGIApplication object that can be run with uvicorn). Most Python ASGI
    frameworks provide an ASGIApplication object. This method also has to
    register all the endpoints, middleware and extensions that are either
    required internally or supplied to it. It must also configure the `startup`
    and `shutdown` methods to be run as part of the ASGI application's lifespan
    or overload the `_run_asgi_app` method to handle the startup and shutdown as
    an alternative.
    * _get_dashboard_endpoints: Gets the dashboard endpoints specs from the
    deployment configuration. Only required if the dashboard files path is set
    in the deployment configuration and the app runner supports serving a
    dashboard alongside the API.
    * _build_cors_middleware: Builds the CORS middleware from the CORS settings
    in the deployment configuration.
    """

    def __init__(
        self, deployment: Union[str, UUID, "DeploymentResponse"], **kwargs: Any
    ):
        """Initialize the deployment app.

        Args:
            deployment: The deployment to run.
            **kwargs: Additional keyword arguments for the deployment app runner.
        """
        self.deployment = self.load_deployment(deployment)
        assert self.deployment.snapshot is not None
        self.snapshot = self.deployment.snapshot

        self.settings = (
            self.snapshot.pipeline_configuration.deployment_settings
        )

        self.service = self.load_deployment_service()

        # Create framework-specific adapters
        self.endpoint_adapter = self._create_endpoint_adapter()
        self.middleware_adapter = self._create_middleware_adapter()
        self._asgi_app: Optional[ASGIApplication] = None

        self.endpoints: List[EndpointSpec] = []
        self.middlewares: List[MiddlewareSpec] = []
        self.extensions: List[AppExtensionSpec] = []

    @property
    def asgi_app(self) -> ASGIApplication:
        """Get the ASGI application.

        Returns:
            The ASGI application.

        Raises:
            RuntimeError: If the ASGI application is not built yet.
        """
        if self._asgi_app is None:
            raise RuntimeError(
                "ASGI application is not built yet. Run the deployment app runner's `build` method first."
            )
        return self._asgi_app

    @classmethod
    def load_deployment(
        cls, deployment: Union[str, UUID, "DeploymentResponse"]
    ) -> DeploymentResponse:
        """Load the deployment.

        Args:
            deployment: The deployment to load.

        Returns:
            The deployment.

        Raises:
            RuntimeError: If the deployment or its snapshot cannot be loaded.
        """
        if isinstance(deployment, str):
            deployment = UUID(deployment)

        if isinstance(deployment, UUID):
            try:
                deployment = Client().zen_store.get_deployment(
                    deployment_id=deployment
                )
            except Exception as e:
                raise RuntimeError(
                    f"Failed to load deployment {deployment}: {e}"
                ) from e
        else:
            assert isinstance(deployment, DeploymentResponse)

        if deployment.snapshot is None:
            raise RuntimeError(f"Deployment {deployment.id} has no snapshot")

        return deployment

    @classmethod
    def load_app_runner(
        cls, deployment: Union[str, UUID, "DeploymentResponse"]
    ) -> "BaseDeploymentAppRunner":
        """Load the app runner for the deployment.

        Args:
            deployment: The deployment to load the app runner for.

        Returns:
            The app runner for the deployment.

        Raises:
            RuntimeError: If the deployment app runner cannot be loaded.
        """
        deployment = cls.load_deployment(deployment)
        assert deployment.snapshot is not None

        settings = (
            deployment.snapshot.pipeline_configuration.deployment_settings
        )

        app_runner_flavor = (
            BaseDeploymentAppRunnerFlavor.load_app_runner_flavor(settings)
        )

        app_runner_cls = app_runner_flavor.implementation_class

        logger.info(
            f"Instantiating deployment app runner class '{app_runner_cls}' for "
            f"deployment {deployment.id}"
        )

        try:
            return app_runner_cls(
                deployment, **settings.deployment_app_runner_kwargs
            )
        except Exception as e:
            raise RuntimeError(
                f"Failed to instantiate deployment app runner class "
                f"'{app_runner_cls}' for deployment {deployment.id}: {e}"
            ) from e

    def load_deployment_service(self) -> BasePipelineDeploymentService:
        """Load the service for the deployment.

        Returns:
            The deployment service for the deployment.

        Raises:
            RuntimeError: If the deployment service cannot be loaded.
        """
        settings = self.snapshot.pipeline_configuration.deployment_settings
        if settings.deployment_service_class is None:
            service_cls: Type[BasePipelineDeploymentService] = (
                PipelineDeploymentService
            )
        else:
            assert isinstance(
                settings.deployment_service_class, SourceOrObject
            )
            try:
                loaded_service_cls = settings.deployment_service_class.load()
            except Exception as e:
                raise RuntimeError(
                    f"Failed to load deployment service from source "
                    f"{settings.deployment_service_class}: {e}\n"
                    "Please check that the source is valid and that the "
                    "deployment service class is importable from the source "
                    "root directory. Hint: run `zenml init` in your local "
                    "source directory to initialize the source root path."
                ) from e

            if not isinstance(loaded_service_cls, type) or not issubclass(
                loaded_service_cls, BasePipelineDeploymentService
            ):
                raise RuntimeError(
                    f"Deployment service class '{loaded_service_cls}' is not a "
                    "subclass of 'BasePipelineDeploymentService'"
                )
            service_cls = loaded_service_cls

        logger.info(
            f"Instantiating deployment service class '{service_cls}' for "
            f"deployment {self.deployment.id}"
        )

        try:
            return service_cls(self, **settings.deployment_service_kwargs)
        except Exception as e:
            raise RuntimeError(
                f"Failed to instantiate deployment service class "
                f"'{service_cls}' for deployment {self.deployment.id}: {e}"
            ) from e

    @property
    @abstractmethod
    def flavor(cls) -> "BaseDeploymentAppRunnerFlavor":
        """Return the flavor associated with this deployment application runner.

        Returns:
            The flavor associated with this deployment application runner.
        """

    @abstractmethod
    def _create_endpoint_adapter(self) -> EndpointAdapter:
        """Create the framework-specific endpoint adapter.

        Returns:
            Endpoint adapter instance for this framework.
        """

    @abstractmethod
    def _create_middleware_adapter(self) -> MiddlewareAdapter:
        """Create the framework-specific middleware adapter.

        Returns:
            Middleware adapter instance for this framework.
        """

    def _build_invoke_endpoint(
        self,
    ) -> Callable[
        [BaseDeploymentInvocationRequest], BaseDeploymentInvocationResponse
    ]:
        """Create the endpoint used to invoke the pipeline deployment.

        Returns:
            The invoke endpoint, built according to the pipeline deployment
            input and output specifications.
        """
        PipelineInvokeRequest, PipelineInvokeResponse = (
            self.service.get_pipeline_invoke_models()
        )

        def _invoke_endpoint(
            request: PipelineInvokeRequest,  # type: ignore[valid-type]
        ) -> PipelineInvokeResponse:  # type: ignore[valid-type]
            return self.service.execute_pipeline(request)

        return _invoke_endpoint

    def dashboard_files_path(self) -> Optional[str]:
        """Get the absolute path of the dashboard files directory.

        Returns:
            Absolute path.

        Raises:
            ValueError: If the dashboard files path is absolute.
            RuntimeError: If the dashboard files path does not exist.
        """
        # If an absolute path is provided, use it
        dashboard_files_path = self.settings.dashboard_files_path
        if not dashboard_files_path:
            import zenml

            return os.path.join(
                zenml.__path__[0], "deployers", "server", "dashboard"
            )

        if os.path.isabs(dashboard_files_path):
            raise ValueError(
                f"Dashboard files path '{dashboard_files_path}' must be "
                "relative to the source root, not absolute."
            )

        # Otherwise, assume this is a path relative to the source root
        source_root = source_utils.get_source_root()
        dashboard_path = os.path.join(source_root, dashboard_files_path)
        if not os.path.exists(dashboard_path):
            raise RuntimeError(
                f"Dashboard files path '{dashboard_path}' does not exist. "
                f"Please check that the path exists and that the source root "
                f"is set correctly. Hint: run `zenml init` in your local source "
                f"directory to initialize the source root path."
            )
        return dashboard_path

    @abstractmethod
    def _get_dashboard_endpoints(self) -> List[EndpointSpec]:
        """Get the dashboard endpoints specs.

        This is called if the dashboard files path is set to construct the
        endpoints specs for the dashboard.

        Returns:
            The dashboard endpoints specs.
        """

    def _create_default_endpoint_specs(self) -> List[EndpointSpec]:
        """Create EndpointSpec objects for default endpoints.

        Returns:
            List of endpoint specs for default endpoints.
        """
        specs = []

        if self.settings.endpoint_enabled(DeploymentDefaultEndpoints.INVOKE):
            specs.append(
                EndpointSpec(
                    path=f"{self.settings.api_url_path}{self.settings.invoke_url_path}",
                    method=EndpointMethod.POST,
                    handler=self._build_invoke_endpoint(),
                    auth_required=True,
                )
            )

        if self.settings.endpoint_enabled(DeploymentDefaultEndpoints.HEALTH):
            specs.append(
                EndpointSpec(
                    path=f"{self.settings.api_url_path}{self.settings.health_url_path}",
                    method=EndpointMethod.GET,
                    handler=self.service.health_check,
                    auth_required=False,
                )
            )

        if self.settings.endpoint_enabled(DeploymentDefaultEndpoints.INFO):
            specs.append(
                EndpointSpec(
                    path=f"{self.settings.api_url_path}{self.settings.info_url_path}",
                    method=EndpointMethod.GET,
                    handler=self.service.get_service_info,
                    auth_required=False,
                )
            )

        if self.settings.endpoint_enabled(DeploymentDefaultEndpoints.METRICS):
            specs.append(
                EndpointSpec(
                    path=f"{self.settings.api_url_path}{self.settings.metrics_url_path}",
                    method=EndpointMethod.GET,
                    handler=self.service.get_execution_metrics,
                    auth_required=False,
                )
            )

        return specs

    def _get_secure_headers(self) -> "Secure":
        """Get the secure headers settings.

        Returns:
            The secure headers settings.
        """
        import secure

        # For each of the secure headers supported by the `secure` library, we
        # check if the corresponding configuration is set in the deployment
        # configuration:
        #
        # - if set to `True`, we use the default value for the header
        # - if set to a string, we use the string as the value for the header
        # - if set to `False`, we don't set the header

        server: Optional[secure.Server] = None
        if self.settings.secure_headers.server:
            server = secure.Server()
            if isinstance(self.settings.secure_headers.server, str):
                server.set(self.settings.secure_headers.server)
            else:
                server.set(str(self.deployment.id))

        hsts: Optional[secure.StrictTransportSecurity] = None
        if self.settings.secure_headers.hsts:
            hsts = secure.StrictTransportSecurity()
            if isinstance(self.settings.secure_headers.hsts, str):
                hsts.set(self.settings.secure_headers.hsts)

        xfo: Optional[secure.XFrameOptions] = None
        if self.settings.secure_headers.xfo:
            xfo = secure.XFrameOptions()
            if isinstance(self.settings.secure_headers.xfo, str):
                xfo.set(self.settings.secure_headers.xfo)

        csp: Optional[secure.ContentSecurityPolicy] = None
        if self.settings.secure_headers.csp:
            csp = secure.ContentSecurityPolicy()
            if isinstance(self.settings.secure_headers.csp, str):
                csp.set(self.settings.secure_headers.csp)

        xcto: Optional[secure.XContentTypeOptions] = None
        if self.settings.secure_headers.content:
            xcto = secure.XContentTypeOptions()
            if isinstance(self.settings.secure_headers.content, str):
                xcto.set(self.settings.secure_headers.content)

        referrer: Optional[secure.ReferrerPolicy] = None
        if self.settings.secure_headers.referrer:
            referrer = secure.ReferrerPolicy()
            if isinstance(self.settings.secure_headers.referrer, str):
                referrer.set(self.settings.secure_headers.referrer)

        cache: Optional[secure.CacheControl] = None
        if self.settings.secure_headers.cache:
            cache = secure.CacheControl()
            if isinstance(self.settings.secure_headers.cache, str):
                cache.set(self.settings.secure_headers.cache)

        permissions: Optional[secure.PermissionsPolicy] = None
        if self.settings.secure_headers.permissions:
            permissions = secure.PermissionsPolicy()
            if isinstance(self.settings.secure_headers.permissions, str):
                # This one is special, because it doesn't allow setting the
                # value as a string, but rather as a list of directives, so we
                # hack our way around it by setting the private _default_value
                # attribute.
                permissions._default_value = (
                    self.settings.secure_headers.permissions
                )

        return secure.Secure(
            server=server,
            hsts=hsts,
            xfo=xfo,
            csp=csp,
            xcto=xcto,
            referrer=referrer,
            cache=cache,
            permissions=permissions,
        )

    def _build_secure_headers_middleware(
        self,
    ) -> MiddlewareSpec:
        """Get the secure headers middleware.

        Returns:
            The secure headers middleware spec.
        """
        secure_headers = self._get_secure_headers()

        async def set_secure_headers(
            app: ASGIApplication,
            scope: Scope,
            receive: ASGIReceiveCallable,
            send: ASGISendCallable,
        ) -> None:
            skip = False
            if scope["type"] != "http":
                skip = True
            else:
                path = scope["path"]

                if path.startswith(
                    self.settings.docs_url_path
                ) or path.startswith(self.settings.redoc_url_path):
                    skip = True

            async def send_wrapper(message: ASGISendEvent) -> None:
                if message["type"] == "http.response.start":
                    hdrs: List[Tuple[bytes, bytes]] = list(
                        message.get("headers", [])
                    )
                    existing = {k: i for i, (k, _) in enumerate(hdrs)}
                    for k, v in secure_headers.headers.items():
                        i = existing.get(k.encode())
                        if i is not None:
                            hdrs[i] = (k.encode(), v.encode())
                        else:
                            hdrs.append((k.encode(), v.encode()))
                    message["headers"] = hdrs
                await send(message)

            wrapped_app = guarantee_single_callable(app)  # type: ignore[no-untyped-call]

            await wrapped_app(scope, receive, send if skip else send_wrapper)

        return MiddlewareSpec(
            middleware=set_secure_headers,
        )

    @abstractmethod
    def _build_cors_middleware(self) -> MiddlewareSpec:
        """Get the CORS middleware.

        Returns:
            The CORS middleware spec.
        """

    def _create_default_middleware_specs(self) -> List[MiddlewareSpec]:
        """Create MiddlewareSpec objects for default middleware.

        Returns:
            List of middleware specs for default middleware.
        """
        specs = []

        if self.settings.middleware_enabled(
            DeploymentDefaultMiddleware.SECURE_HEADERS
        ):
            specs.append(self._build_secure_headers_middleware())

        if self.settings.middleware_enabled(DeploymentDefaultMiddleware.CORS):
            specs.append(self._build_cors_middleware())

        return specs

    def install_extensions(self, *extension_specs: AppExtensionSpec) -> None:
        """Install the given app extensions.

        Args:
            extension_specs: The app extensions to install.

        Raises:
            ValueError: If the extension is not a subclass of BaseAppExtension.
            RuntimeError: If the extension cannot be initialized.
        """
        for ext_spec in extension_specs:
            # Load extension
            ext_spec.load_sources()
            extension_obj = ext_spec.resolve_extension_handler()

            # Handle callable vs class-based extensions
            if isinstance(extension_obj, type):
                if not issubclass(extension_obj, BaseAppExtension):
                    raise ValueError(
                        f"Extension type {extension_obj} is not a subclass of "
                        "BaseAppExtension"
                    )

                try:
                    extension_instance = extension_obj(
                        **ext_spec.extension_kwargs
                    )
                except Exception as e:
                    raise RuntimeError(
                        f"Failed to initialize extension class {extension_obj}: {e}"
                    ) from e

                extension_instance.install(self)
            else:
                # Simple callable extension
                extension_obj(
                    app_runner=self,
                    **ext_spec.extension_kwargs,
                )
            self.extensions.append(ext_spec)

    def register_endpoints(self, *endpoint_specs: EndpointSpec) -> None:
        """Register the given endpoints.

        Args:
            endpoint_specs: The endpoints to register.
        """
        for endpoint_spec in endpoint_specs:
            endpoint_spec.load_sources()
            self.endpoint_adapter.register_endpoint(self, endpoint_spec)
            self.endpoints.append(endpoint_spec)

    def register_middlewares(self, *middleware_specs: MiddlewareSpec) -> None:
        """Register the given middleware.

        Args:
            middleware_specs: The middleware to register.
        """
        for middleware_spec in middleware_specs:
            middleware_spec.load_sources()
            self.middleware_adapter.register_middleware(self, middleware_spec)
            self.middlewares.append(middleware_spec)

    def _run_startup_hook(self) -> None:
        """Run the startup hook.

        Raises:
            ValueError: If the startup hook is not callable.
            Exception: If the startup hook fails to execute.
        """
        if not self.settings.startup_hook:
            return

        assert isinstance(self.settings.startup_hook, SourceOrObject)
        startup_hook = self.settings.startup_hook.load()

        if not callable(startup_hook):
            raise ValueError(
                f"The startup hook object {startup_hook} must be callable"
            )

        logger.info("Executing the deployment application startup hook...")
        try:
            startup_hook(
                app_runner=self,
                **self.settings.startup_hook_kwargs,
            )
        except Exception as e:
            logger.exception(f"Failed to execute startup hook: {e}")
            raise

    def startup(self) -> None:
        """Startup the deployment app.

        Raises:
            Exception: If the service initialization fails.
        """
        logger.info("🚀 Initializing the pipeline deployment service...")

        try:
            self.service.initialize()
            logger.info(
                "✅ Pipeline deployment service initialized successfully"
            )
        except Exception as e:
            logger.error(
                f"❌ Failed to initialize the pipeline deployment service: {e}"
            )
            raise

        self._run_startup_hook()

    def _run_shutdown_hook(self) -> None:
        """Run the shutdown hook.

        Raises:
            ValueError: If the shutdown hook is not callable.
            Exception: If the shutdown hook fails to execute.
        """
        if not self.settings.shutdown_hook:
            return

        assert isinstance(self.settings.shutdown_hook, SourceOrObject)

        shutdown_hook = self.settings.shutdown_hook.load()

        if not shutdown_hook:
            return

        if not callable(shutdown_hook):
            raise ValueError(
                f"The shutdown hook object {shutdown_hook} must be callable"
            )

        logger.info("Executing the deployment application shutdown hook...")
        try:
            shutdown_hook(
                app_runner=self,
                **self.settings.shutdown_hook_kwargs,
            )
        except Exception as e:
            logger.exception(f"Failed to execute shutdown hook: {e}")
            raise

    def shutdown(self) -> None:
        """Shutdown the deployment app.

        Raises:
            Exception: If the service cleanup fails.
        """
        self._run_shutdown_hook()

        logger.info("🛑 Cleaning up the pipeline deployment service...")
        try:
            self.service.cleanup()
            logger.info(
                "✅ The pipeline deployment service was cleaned up successfully"
            )
        except Exception as e:
            logger.error(
                f"❌ Failed to clean up the pipeline deployment service: {e}"
            )
            raise

    def build_asgi_app(self) -> ASGIApplication:
        """Build the ASGI application.

        Returns:
            The ASGI application.
        """
        endpoints = self._create_default_endpoint_specs()

        custom_endpoints = (
            self.settings.custom_endpoints
            if self.settings.custom_endpoints
            else []
        )

        # Allow custom endpoints to override default endpoints when they share
        # the same path and method.

        def normalize_path(path: str) -> str:
            if path and path != "/":
                path = re.sub(r"(?<!^)/+$", "", path)

            # normalize typed path params: {name:path} -> {param:path}
            path = re.sub(r"\{[^{}:/]+:path\}", "{param:path}", path)

            # normalize untyped params: {name} -> {param}
            path = re.sub(r"\{[^{}:/]+\}", "{param}", path)

            return path

        custom_keys = {
            (e.method, normalize_path(e.path)) for e in custom_endpoints
        }
        endpoints = [
            e
            for e in endpoints
            if (e.method, normalize_path(e.path)) not in custom_keys
        ]

        endpoints.extend(custom_endpoints)

        if self.settings.endpoint_enabled(
            DeploymentDefaultEndpoints.DASHBOARD
        ):
            endpoints.extend(self._get_dashboard_endpoints())

        middlewares = self._create_default_middleware_specs()

        if self.settings.custom_middlewares:
            middlewares.extend(self.settings.custom_middlewares)

        extensions = []
        if self.settings.app_extensions:
            extensions.extend(self.settings.app_extensions)

        return self.build(middlewares, endpoints, extensions)

    def run(self) -> None:
        """Run the ASGI application via uvicorn.

        Raises:
            Exception: If the application fails to start.
        """
        import uvicorn

        settings = self.settings

        logger.info(f"""
🚀 Starting ZenML pipeline deployment application:
   Deployment ID: {self.deployment.id}
   Deployment Name: {self.deployment.name}
   Snapshot ID: {self.snapshot.id}
   Snapshot Name: {self.snapshot.name or "N/A"}
   Pipeline ID: {self.snapshot.pipeline.id}
   Pipeline Name: {self.snapshot.pipeline.name}
   Host: {settings.uvicorn_host}
   Port: {settings.uvicorn_port}
   Workers: {settings.uvicorn_workers}
   Log Level: {settings.log_level}
""")

        uvicorn_kwargs: Dict[str, Any] = dict(
            host=settings.uvicorn_host,
            port=settings.uvicorn_port,
            workers=settings.uvicorn_workers,
            log_level=settings.log_level.value,
            reload=settings.uvicorn_reload,
            access_log=True,
        )
        if settings.uvicorn_kwargs:
            uvicorn_kwargs.update(settings.uvicorn_kwargs)

        if settings.uvicorn_workers > 1 or settings.uvicorn_reload:
            # These settings require the app to run as a subprocess, so we need
            # to use the factory pattern and pass the app as a source object.
            os.environ[ZENML_DEPLOYMENT_ID_ENV_VAR] = str(self.deployment.id)

            app_source = source_utils.resolve(build_asgi_app)
            app_path = f"{app_source.module}:{app_source.attribute}"
            uvicorn_kwargs.update(dict(app=app_path, factory=True))
        else:
            # The regular path is to build the app and pass it as an object in
            # the same process.
            app = self.build_asgi_app()
            uvicorn_kwargs.update(dict(app=app))

        try:
            # Start the ASGI application
            uvicorn.run(
                **uvicorn_kwargs,
            )
        except KeyboardInterrupt:
            pass
        except Exception as e:
            logger.error(
                f"❌ Failed to start deployment application: {str(e)}"
            )
            raise
        logger.info("\n🛑 Deployment application shutdown")

    @abstractmethod
    def build(
        self,
        middlewares: List[MiddlewareSpec],
        endpoints: List[EndpointSpec],
        extensions: List[AppExtensionSpec],
    ) -> ASGIApplication:
        """Build the ASGI compatible web application.

        Args:
            middlewares: The middleware to register.
            endpoints: The endpoints to register.
            extensions: The extensions to install.

        Returns:
            The ASGI compatible web application.
        """


class BaseDeploymentAppRunnerFlavor(ABC):
    """Base class for deployment app runner flavors.

    BaseDeploymentAppRunner implementations must also provide implementations
    for this class. The flavor class implementation should be kept separate from
    the implementation class to allow it to be imported without importing the
    implementation class and all its dependencies.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """The name of the deployment app runner flavor.

        Returns:
            The name of the deployment app runner flavor.
        """

    @property
    @abstractmethod
    def implementation_class(self) -> Type[BaseDeploymentAppRunner]:
        """The class that implements the deployment app runner.

        Returns:
            The implementation class for the deployment app runner.
        """

    @property
    def requirements(self) -> List[str]:
        """The software requirements for the deployment app runner.

        Returns:
            The software requirements for the deployment app runner.
        """
        return ["uvicorn", "secure~=1.0.1", "asgiref~=3.10.0", "Jinja2"]

    @classmethod
    def load_app_runner_flavor(
        cls, settings: DeploymentSettings
    ) -> "BaseDeploymentAppRunnerFlavor":
        """Load the app runner flavor for the deployment settings.

        Args:
            settings: The deployment settings to load the app runner flavor for.

        Returns:
            The app runner flavor for the deployment.

        Raises:
            RuntimeError: If the deployment app runner flavor cannot be loaded.
        """
        from zenml.deployers.server.fastapi import (
            FastAPIDeploymentAppRunnerFlavor,
        )

        if settings.deployment_app_runner_flavor is None:
            app_runner_flavor_class: Type[BaseDeploymentAppRunnerFlavor] = (
                FastAPIDeploymentAppRunnerFlavor
            )
        else:
            assert isinstance(
                settings.deployment_app_runner_flavor, SourceOrObject
            )
            try:
                loaded_app_runner_flavor_class = (
                    settings.deployment_app_runner_flavor.load()
                )
            except Exception as e:
                raise RuntimeError(
                    f"Failed to load deployment app runner flavor from source "
                    f"{settings.deployment_app_runner_flavor}: {e}\n"
                    "Please check that the source is valid and that the "
                    "deployment app runner flavor class is importable from the "
                    "source root directory. Hint: run `zenml init` in your "
                    "local source directory to initialize the source root path."
                ) from e

            if not isinstance(
                loaded_app_runner_flavor_class, type
            ) or not issubclass(
                loaded_app_runner_flavor_class, BaseDeploymentAppRunnerFlavor
            ):
                raise RuntimeError(
                    f"The object '{loaded_app_runner_flavor_class}' is not a "
                    "subclass of 'BaseDeploymentAppRunnerFlavor'"
                )

            app_runner_flavor_class = loaded_app_runner_flavor_class

        try:
            app_runner_flavor = app_runner_flavor_class()
        except Exception as e:
            raise RuntimeError(
                f"Failed to instantiate deployment app runner flavor "
                f"'{loaded_app_runner_flavor_class}': {e}"
            ) from e

        return app_runner_flavor


def build_asgi_app() -> ASGIApplication:
    """Build the ASGI application.

    This is the factory method called by uvicorn to build the ASGI
    application. It is not allowed to receive any arguments and therefore
    the deployment ID must be passed in as the ZENML_DEPLOYMENT_ID
    environment variable.

    Returns:
        The ASGI application.

    Raises:
        RuntimeError: If the deployment ID is not set in the
            ZENML_DEPLOYMENT_ID environment variable.
    """
    deployment_id = os.getenv(ZENML_DEPLOYMENT_ID_ENV_VAR)
    if not deployment_id:
        raise RuntimeError(
            f"{ZENML_DEPLOYMENT_ID_ENV_VAR} environment variable is not set"
        )

    app_runner = BaseDeploymentAppRunner.load_app_runner(deployment_id)
    return app_runner.build_asgi_app()


def start_deployment_app(
    deployment_id: UUID,
    pid_file: Optional[str] = None,
    log_file: Optional[str] = None,
    host: Optional[str] = None,
    port: Optional[int] = None,
    reload: Optional[bool] = None,
) -> None:
    """Start the deployment app.

    Args:
        deployment_id: The deployment ID.
        pid_file: The PID file to use for the deployment.
        log_file: The log file to use for the deployment.
        host: The custom host to use for the deployment.
        port: The custom port to use for the deployment.
        reload: Whether to automatically reload the deployment when the
            code changes.
    """
    if pid_file or log_file:
        # create parent directory if necessary
        for f in (pid_file, log_file):
            if f:
                os.makedirs(os.path.dirname(f), exist_ok=True)

        setup_daemon(pid_file, log_file)

    logger.info(
        f"Starting deployment application server for deployment "
        f"{deployment_id}"
    )

    app_runner = BaseDeploymentAppRunner.load_app_runner(deployment_id)

    # Allow host/port overrides coming from the CLI.
    if host:
        app_runner.settings.uvicorn_host = host
    if port:
        app_runner.settings.uvicorn_port = int(port)
    if reload is not None:
        app_runner.settings.uvicorn_reload = reload
    app_runner.run()


if __name__ == "__main__":
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--deployment_id",
        default=os.getenv(ZENML_DEPLOYMENT_ID_ENV_VAR),
        help="Pipeline snapshot ID",
    )
    parser.add_argument(
        "--pid_file",
        default=os.getenv("ZENML_PID_FILE"),
        help="PID file to use for the deployment",
    )
    parser.add_argument(
        "--log_file",
        default=os.getenv("ZENML_LOG_FILE"),
        help="Log file to use for the deployment",
    )
    parser.add_argument(
        "--host",
        default=os.getenv("ZENML_UVICORN_HOST"),
        help=(
            "Optional host override for the uvicorn server. If provided, "
            "this takes precedence over the deployment settings."
        ),
    )
    parser.add_argument(
        "--port",
        type=int,
        default=os.getenv("ZENML_UVICORN_PORT"),
        help=(
            "Optional port override for the uvicorn server. If provided, "
            "this takes precedence over the deployment settings."
        ),
    )
    parser.add_argument(
        "--reload",
        type=bool,
        default=os.getenv("ZENML_UVICORN_RELOAD"),
        help=(
            "Whether to automatically reload the deployment when the code changes."
        ),
    )
    args = parser.parse_args()

    # Activate integrations to ensure all components are available
    integration_registry.activate_integrations()

    start_deployment_app(
        deployment_id=args.deployment_id,
        pid_file=args.pid_file,
        log_file=args.log_file,
        host=args.host,
        port=args.port,
        reload=args.reload,
    )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""SQLModel implementation of step run tables."""

import json
from datetime import datetime
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence
from uuid import UUID

from pydantic import ConfigDict
from sqlalchemy import TEXT, Column, String, UniqueConstraint
from sqlalchemy.dialects.mysql import MEDIUMTEXT
from sqlalchemy.orm import joinedload, selectinload
from sqlalchemy.sql.base import ExecutableOption
from sqlmodel import Field, Relationship, SQLModel

from zenml.config.pipeline_configurations import PipelineConfiguration
from zenml.config.step_configurations import Step
from zenml.constants import MEDIUMTEXT_MAX_LENGTH
from zenml.enums import (
    ExecutionStatus,
    MetadataResourceTypes,
    PipelineRunTriggeredByType,
    StepRunInputArtifactType,
)
from zenml.models import (
    ExceptionInfo,
    StepRunRequest,
    StepRunResponse,
    StepRunResponseBody,
    StepRunResponseMetadata,
    StepRunUpdate,
)
from zenml.models.v2.core.artifact_version import ArtifactVersionResponse
from zenml.models.v2.core.step_run import (
    StepRunInputResponse,
    StepRunResponseResources,
)
from zenml.utils.time_utils import utc_now
from zenml.zen_stores.schemas.base_schemas import NamedSchema
from zenml.zen_stores.schemas.constants import MODEL_VERSION_TABLENAME
from zenml.zen_stores.schemas.pipeline_run_schemas import PipelineRunSchema
from zenml.zen_stores.schemas.pipeline_snapshot_schemas import (
    PipelineSnapshotSchema,
    StepConfigurationSchema,
)
from zenml.zen_stores.schemas.project_schemas import ProjectSchema
from zenml.zen_stores.schemas.schema_utils import (
    build_foreign_key_field,
    build_index,
)
from zenml.zen_stores.schemas.user_schemas import UserSchema
from zenml.zen_stores.schemas.utils import (
    RunMetadataInterface,
    jl_arg,
)

if TYPE_CHECKING:
    from zenml.zen_stores.schemas.artifact_schemas import ArtifactVersionSchema
    from zenml.zen_stores.schemas.logs_schemas import LogsSchema
    from zenml.zen_stores.schemas.model_schemas import ModelVersionSchema
    from zenml.zen_stores.schemas.run_metadata_schemas import RunMetadataSchema


class StepRunSchema(NamedSchema, RunMetadataInterface, table=True):
    """SQL Model for steps of pipeline runs."""

    __tablename__ = "step_run"
    __table_args__ = (
        UniqueConstraint(
            "name",
            "pipeline_run_id",
            "version",
            name="unique_step_name_for_pipeline_run",
        ),
        build_index(
            table_name=__tablename__,
            column_names=[
                "cache_key",
            ],
        ),
    )

    # Fields
    start_time: Optional[datetime] = Field(nullable=True)
    end_time: Optional[datetime] = Field(nullable=True)
    latest_heartbeat: Optional[datetime] = Field(
        nullable=True,
        description="The latest execution heartbeat.",
    )
    status: str = Field(nullable=False)

    docstring: Optional[str] = Field(sa_column=Column(TEXT, nullable=True))
    cache_key: Optional[str] = Field(nullable=True)
    cache_expires_at: Optional[datetime] = Field(nullable=True)
    source_code: Optional[str] = Field(sa_column=Column(TEXT, nullable=True))
    code_hash: Optional[str] = Field(nullable=True)
    version: int = Field(nullable=False)
    is_retriable: bool = Field(nullable=False)

    exception_info: Optional[str] = Field(
        sa_column=Column(
            String(length=MEDIUMTEXT_MAX_LENGTH).with_variant(
                MEDIUMTEXT, "mysql"
            ),
            nullable=True,
        )
    )
    heartbeat_threshold: Optional[int] = Field(nullable=True)
    # Foreign keys
    original_step_run_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=__tablename__,
        source_column="original_step_run_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    snapshot_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=PipelineSnapshotSchema.__tablename__,
        source_column="snapshot_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=True,
    )
    pipeline_run_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=PipelineRunSchema.__tablename__,
        source_column="pipeline_run_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    user_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=UserSchema.__tablename__,
        source_column="user_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )
    project_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=ProjectSchema.__tablename__,
        source_column="project_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
    )
    model_version_id: Optional[UUID] = build_foreign_key_field(
        source=__tablename__,
        target=MODEL_VERSION_TABLENAME,
        source_column="model_version_id",
        target_column="id",
        ondelete="SET NULL",
        nullable=True,
    )

    # Relationships
    project: "ProjectSchema" = Relationship(back_populates="step_runs")
    user: Optional["UserSchema"] = Relationship(back_populates="step_runs")
    snapshot: Optional["PipelineSnapshotSchema"] = Relationship(
        back_populates="step_runs"
    )
    run_metadata: List["RunMetadataSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            secondary="run_metadata_resource",
            primaryjoin=f"and_(foreign(RunMetadataResourceSchema.resource_type)=='{MetadataResourceTypes.STEP_RUN.value}', foreign(RunMetadataResourceSchema.resource_id)==StepRunSchema.id)",
            secondaryjoin="RunMetadataSchema.id==foreign(RunMetadataResourceSchema.run_metadata_id)",
            overlaps="run_metadata",
        ),
    )
    input_artifacts: List["StepRunInputArtifactSchema"] = Relationship(
        sa_relationship_kwargs={"cascade": "delete"}
    )
    output_artifacts: List["StepRunOutputArtifactSchema"] = Relationship(
        sa_relationship_kwargs={"cascade": "delete"}
    )
    logs: List["LogsSchema"] = Relationship(
        back_populates="step_run",
        sa_relationship_kwargs={"cascade": "delete"},
    )
    parents: List["StepRunParentsSchema"] = Relationship(
        sa_relationship_kwargs={
            "cascade": "delete",
            "primaryjoin": "StepRunParentsSchema.child_id == StepRunSchema.id",
        },
    )
    pipeline_run: "PipelineRunSchema" = Relationship(
        back_populates="step_runs"
    )
    model_version: "ModelVersionSchema" = Relationship(
        back_populates="step_runs",
    )
    triggered_runs: List["PipelineRunSchema"] = Relationship(
        sa_relationship_kwargs={
            "viewonly": True,
            "primaryjoin": f"and_(foreign(PipelineRunSchema.triggered_by) == StepRunSchema.id, foreign(PipelineRunSchema.triggered_by_type) == '{PipelineRunTriggeredByType.STEP_RUN.value}')",
        },
    )

    original_step_run: Optional["StepRunSchema"] = Relationship(
        sa_relationship_kwargs={"remote_side": "StepRunSchema.id"}
    )
    # In static pipelines, we use the config that is compiled in the snapshot.
    static_config: Optional["StepConfigurationSchema"] = Relationship(
        sa_relationship_kwargs=dict(
            viewonly=True,
            primaryjoin="and_(foreign(StepConfigurationSchema.name) == StepRunSchema.name, foreign(StepConfigurationSchema.snapshot_id) == StepRunSchema.snapshot_id)",
        ),
    )
    # In dynamic pipelines, the config is dynamically generated and cannot be
    # included in the compiled snapshot. In this case, we link it directly to
    # the step run.
    dynamic_config: Optional["StepConfigurationSchema"] = Relationship(
        sa_relationship_kwargs={
            "cascade": "all, delete-orphan",
        },
    )
    # In legacy pipelines (before snapshots, former deployments), the config
    # is stored as a string in the step run.
    step_configuration: str = Field(
        sa_column=Column(
            String(length=MEDIUMTEXT_MAX_LENGTH).with_variant(
                MEDIUMTEXT, "mysql"
            ),
            nullable=True,
        )
    )

    model_config = ConfigDict(protected_namespaces=())  # type: ignore[assignment]

    @classmethod
    def get_query_options(
        cls,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> Sequence[ExecutableOption]:
        """Get the query options for the schema.

        Args:
            include_metadata: Whether metadata will be included when converting
                the schema to a model.
            include_resources: Whether resources will be included when
                converting the schema to a model.
            **kwargs: Keyword arguments to allow schema specific logic

        Returns:
            A list of query options.
        """
        from zenml.zen_stores.schemas import (
            ArtifactVersionSchema,
            ModelVersionSchema,
        )

        options = [
            selectinload(jl_arg(StepRunSchema.snapshot)).load_only(
                jl_arg(PipelineSnapshotSchema.pipeline_configuration)
            ),
            selectinload(jl_arg(StepRunSchema.pipeline_run)).load_only(
                jl_arg(PipelineRunSchema.start_time)
            ),
            joinedload(jl_arg(StepRunSchema.static_config)),
            joinedload(jl_arg(StepRunSchema.dynamic_config)),
        ]

        # if include_metadata:
        #     options.extend(
        #         [
        #             joinedload(jl_arg(StepRunSchema.parents)),
        #             joinedload(jl_arg(StepRunSchema.run_metadata)),
        #         ]
        #     )

        if include_resources:
            options.extend(
                [
                    selectinload(
                        jl_arg(StepRunSchema.model_version)
                    ).joinedload(
                        jl_arg(ModelVersionSchema.model), innerjoin=True
                    ),
                    selectinload(jl_arg(StepRunSchema.user)),
                    selectinload(jl_arg(StepRunSchema.input_artifacts))
                    .joinedload(
                        jl_arg(StepRunInputArtifactSchema.artifact_version),
                        innerjoin=True,
                    )
                    .joinedload(
                        jl_arg(ArtifactVersionSchema.artifact), innerjoin=True
                    ),
                    selectinload(jl_arg(StepRunSchema.output_artifacts))
                    .joinedload(
                        jl_arg(StepRunOutputArtifactSchema.artifact_version),
                        innerjoin=True,
                    )
                    .joinedload(
                        jl_arg(ArtifactVersionSchema.artifact), innerjoin=True
                    ),
                    selectinload(jl_arg(StepRunSchema.logs)),
                ]
            )

        return options

    @classmethod
    def from_request(
        cls,
        request: StepRunRequest,
        snapshot_id: Optional[UUID],
        version: int,
        is_retriable: bool,
    ) -> "StepRunSchema":
        """Create a step run schema from a step run request model.

        Args:
            request: The step run request model.
            snapshot_id: The snapshot ID.
            version: The version of the step run.
            is_retriable: Whether the step run is retriable.

        Returns:
            The step run schema.
        """
        return cls(
            name=request.name,
            project_id=request.project,
            user_id=request.user,
            start_time=request.start_time,
            end_time=request.end_time,
            status=request.status.value,
            snapshot_id=snapshot_id,
            original_step_run_id=request.original_step_run_id,
            pipeline_run_id=request.pipeline_run_id,
            docstring=request.docstring,
            cache_key=request.cache_key,
            cache_expires_at=request.cache_expires_at,
            code_hash=request.code_hash,
            source_code=request.source_code,
            version=version,
            is_retriable=is_retriable,
            exception_info=request.exception_info.model_dump_json()
            if request.exception_info
            else None,
        )

    def get_step_configuration(self) -> Step:
        """Get the step configuration for the step run.

        Raises:
            ValueError: If the step run has no step configuration.

        Returns:
            The step configuration.
        """
        step = None

        if self.snapshot is not None:
            if config_schema := (self.dynamic_config or self.static_config):
                pipeline_configuration = (
                    PipelineConfiguration.model_validate_json(
                        self.snapshot.pipeline_configuration
                    )
                )
                pipeline_configuration.finalize_substitutions(
                    start_time=self.pipeline_run.start_time,
                    inplace=True,
                )
                step = Step.from_dict(
                    json.loads(config_schema.config),
                    pipeline_configuration=pipeline_configuration,
                )
        if not step and self.step_configuration:
            # In this legacy case, we're guaranteed to have the merged
            # config stored in the DB, which means we can instantiate the
            # `Step` object directly without passing the pipeline
            # configuration.
            step = Step.model_validate_json(self.step_configuration)
        elif not step:
            raise ValueError(
                f"Unable to load the configuration for step `{self.name}` from "
                "the database. To solve this please delete the pipeline run "
                "that this step run belongs to. Pipeline Run ID: "
                f"`{self.pipeline_run_id}`."
            )

        return step

    def to_model(
        self,
        include_metadata: bool = False,
        include_resources: bool = False,
        **kwargs: Any,
    ) -> StepRunResponse:
        """Convert a `StepRunSchema` to a `StepRunResponse`.

        Args:
            include_metadata: Whether the metadata will be filled.
            include_resources: Whether the resources will be filled.
            **kwargs: Keyword arguments to allow schema specific logic


        Returns:
            The created StepRunResponse.
        """
        step = self.get_step_configuration()

        body = StepRunResponseBody(
            user_id=self.user_id,
            project_id=self.project_id,
            type=step.config.step_type,
            status=ExecutionStatus(self.status),
            version=self.version,
            is_retriable=self.is_retriable,
            start_time=self.start_time,
            end_time=self.end_time,
            latest_heartbeat=self.latest_heartbeat,
            created=self.created,
            updated=self.updated,
            model_version_id=self.model_version_id,
            substitutions=step.config.substitutions,
            heartbeat_threshold=self.heartbeat_threshold,
        )
        metadata = None
        if include_metadata:
            metadata = StepRunResponseMetadata(
                config=step.config,
                spec=step.spec,
                cache_key=self.cache_key,
                cache_expires_at=self.cache_expires_at,
                code_hash=self.code_hash,
                docstring=self.docstring,
                source_code=self.source_code,
                exception_info=ExceptionInfo.model_validate_json(
                    self.exception_info
                )
                if self.exception_info
                else None,
                snapshot_id=self.snapshot_id,
                pipeline_run_id=self.pipeline_run_id,
                original_step_run_id=self.original_step_run_id,
                parent_step_ids=[p.parent_id for p in self.parents],
                run_metadata=self.fetch_metadata(),
            )

        resources = None
        if include_resources:
            model_version = None
            if self.model_version:
                model_version = self.model_version.to_model()

            input_artifacts: Dict[str, List[StepRunInputResponse]] = {}
            for input_artifact in self.input_artifacts:
                if input_artifact.name not in input_artifacts:
                    input_artifacts[input_artifact.name] = []
                step_run_input = StepRunInputResponse(
                    input_type=StepRunInputArtifactType(input_artifact.type),
                    index=input_artifact.input_index,
                    chunk_index=input_artifact.chunk_index,
                    chunk_size=input_artifact.chunk_size,
                    **input_artifact.artifact_version.to_model().model_dump(),
                )
                input_artifacts[input_artifact.name].append(step_run_input)

            for artifact_list in input_artifacts.values():
                artifact_list.sort(key=lambda a: a.index or 0)

            output_artifacts: Dict[str, List["ArtifactVersionResponse"]] = {}
            for output_artifact in self.output_artifacts:
                if output_artifact.name not in output_artifacts:
                    output_artifacts[output_artifact.name] = []
                output_artifacts[output_artifact.name].append(
                    output_artifact.artifact_version.to_model()
                )

            resources = StepRunResponseResources(
                user=self.user.to_model() if self.user else None,
                model_version=model_version,
                log_collection=[log.to_model() for log in self.logs],
                inputs=input_artifacts,
                outputs=output_artifacts,
            )

        return StepRunResponse(
            id=self.id,
            name=self.name,
            body=body,
            metadata=metadata,
            resources=resources,
        )

    def update(self, step_update: "StepRunUpdate") -> "StepRunSchema":
        """Update a step run schema with a step run update model.

        Args:
            step_update: The step run update model.

        Returns:
            The updated step run schema.
        """
        for key, value in step_update.model_dump(
            exclude_unset=True, exclude_none=True
        ).items():
            if key == "status":
                self.status = value.value
            if key == "end_time":
                self.end_time = value
            if key == "exception_info":
                self.exception_info = json.dumps(value)
            if key == "cache_expires_at":
                self.cache_expires_at = value

        self.updated = utc_now()

        return self


class StepRunParentsSchema(SQLModel, table=True):
    """SQL Model that defines the order of steps."""

    __tablename__ = "step_run_parents"

    # Foreign Keys
    parent_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=StepRunSchema.__tablename__,
        source_column="parent_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )
    child_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=StepRunSchema.__tablename__,
        source_column="child_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )


class StepRunInputArtifactSchema(SQLModel, table=True):
    """SQL Model that defines which artifacts are inputs to which step."""

    __tablename__ = "step_run_input_artifact"

    # Fields
    name: str = Field(nullable=False, primary_key=True)
    type: str
    input_index: int = Field(nullable=False, primary_key=True)
    chunk_index: Optional[int] = None
    chunk_size: Optional[int] = None

    # Foreign keys
    step_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=StepRunSchema.__tablename__,
        source_column="step_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )
    # Note: We keep the name artifact_id instead of artifact_version_id here to
    # avoid having to drop and recreate the primary key constraint.
    artifact_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target="artifact_version",
        source_column="artifact_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )

    # Relationships
    step_run: "StepRunSchema" = Relationship(back_populates="input_artifacts")
    artifact_version: "ArtifactVersionSchema" = Relationship(
        back_populates="input_of_step_runs",
        sa_relationship_kwargs={"lazy": "joined"},
    )


class StepRunOutputArtifactSchema(SQLModel, table=True):
    """SQL Model that defines which artifacts are outputs of which step."""

    __tablename__ = "step_run_output_artifact"

    # Fields
    name: str

    # Foreign keys
    step_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target=StepRunSchema.__tablename__,
        source_column="step_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )
    # Note: we keep the name artifact_id instead of artifact_version_id here to
    # avoid having to drop and recreate the primary key constraint.
    artifact_id: UUID = build_foreign_key_field(
        source=__tablename__,
        target="artifact_version",
        source_column="artifact_id",
        target_column="id",
        ondelete="CASCADE",
        nullable=False,
        primary_key=True,
    )

    # Relationship
    step_run: "StepRunSchema" = Relationship(back_populates="output_artifacts")
    artifact_version: "ArtifactVersionSchema" = Relationship(
        back_populates="output_of_step_runs",
        sa_relationship_kwargs={"lazy": "joined"},
    )



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Implementation of the ZenML local orchestrator."""

import time
from typing import TYPE_CHECKING, Dict, List, Optional, Type
from uuid import uuid4

from zenml.enums import ExecutionMode
from zenml.logger import get_logger
from zenml.orchestrators import (
    BaseOrchestrator,
    BaseOrchestratorConfig,
    BaseOrchestratorFlavor,
    SubmissionResult,
)
from zenml.stack import Stack
from zenml.utils import string_utils
from zenml.utils.env_utils import temporary_environment

if TYPE_CHECKING:
    from zenml.models import PipelineRunResponse, PipelineSnapshotResponse

logger = get_logger(__name__)


class LocalOrchestrator(BaseOrchestrator):
    """Orchestrator responsible for running pipelines locally.

    This orchestrator does not allow for concurrent execution of steps and also
    does not support running on a schedule.
    """

    _orchestrator_run_id: Optional[str] = None

    @property
    def run_init_cleanup_at_step_level(self) -> bool:
        """Whether the orchestrator runs the init and cleanup hooks at step level.

        For orchestrators that run their steps in isolated step environments,
        the run context cannot be shared between steps. In this case, the init
        and cleanup hooks need to be run at step level for each individual step.

        For orchestrators that run their steps in a shared environment with a
        shared memory (e.g. the local orchestrator), the init and cleanup hooks
        can be run at run level and this property should be overridden to return
        True.

        Returns:
            Whether the orchestrator runs the init and cleanup hooks at step
            level.
        """
        return False

    def submit_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        base_environment: Dict[str, str],
        step_environments: Dict[str, Dict[str, str]],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a pipeline to the orchestrator.

        This method should only submit the pipeline and not wait for it to
        complete. If the orchestrator is configured to wait for the pipeline run
        to complete, a function that waits for the pipeline run to complete can
        be passed as part of the submission result.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            base_environment: Base environment shared by all steps. This should
                be set if your orchestrator for example runs one container that
                is responsible for starting all the steps.
            step_environments: Environment variables to set when executing
                specific steps.
            placeholder_run: An optional placeholder run for the snapshot.

        Returns:
            Optional submission result.

        Raises:
            step_exception: The exception that occurred while running a failed
                step.
            RuntimeError: If the pipeline run fails.
        """
        self._orchestrator_run_id = str(uuid4())
        start_time = time.time()

        execution_mode = snapshot.pipeline_configuration.execution_mode

        failed_steps: List[str] = []
        step_exception: Optional[Exception] = None
        skipped_steps: List[str] = []

        self.run_init_hook(snapshot=snapshot)

        # Run each step
        for step_name, step in snapshot.step_configurations.items():
            if (
                execution_mode == ExecutionMode.STOP_ON_FAILURE
                and failed_steps
            ):
                logger.warning(
                    "Skipping step %s due to the failed step(s): %s (Execution mode %s)",
                    step_name,
                    ", ".join(failed_steps),
                    execution_mode,
                )
                skipped_steps.append(step_name)
                continue

            if failed_upstream_steps := [
                fs for fs in failed_steps if fs in step.spec.upstream_steps
            ]:
                logger.warning(
                    "Skipping step %s due to failure in upstream step(s): %s (Execution mode %s)",
                    step_name,
                    ", ".join(failed_upstream_steps),
                    execution_mode,
                )
                skipped_steps.append(step_name)
                continue

            if skipped_upstream_steps := [
                fs for fs in skipped_steps if fs in step.spec.upstream_steps
            ]:
                logger.warning(
                    "Skipping step %s due to the skipped upstream step(s) %s (Execution mode %s)",
                    step_name,
                    ", ".join(skipped_upstream_steps),
                    execution_mode,
                )
                skipped_steps.append(step_name)
                continue

            if self.requires_resources_in_orchestration_environment(step):
                logger.warning(
                    "Specifying step resources is not supported for the local "
                    "orchestrator, ignoring resource configuration for "
                    "step %s.",
                    step_name,
                )

            step_environment = step_environments[step_name]
            try:
                with temporary_environment(step_environment):
                    self.run_step(step=step)
            except Exception as e:
                failed_steps.append(step_name)
                logger.exception("Step %s failed.", step_name)

                if execution_mode == ExecutionMode.FAIL_FAST:
                    step_exception = e
                    break

        self.run_cleanup_hook(snapshot=snapshot)

        if execution_mode == ExecutionMode.FAIL_FAST and failed_steps:
            assert step_exception is not None
            raise step_exception

        if failed_steps:
            raise RuntimeError(
                "Pipeline run has failed due to failure in step(s): "
                f"{', '.join(failed_steps)}"
            )

        run_duration = time.time() - start_time
        logger.info(
            "Pipeline run has finished in `%s`.",
            string_utils.get_human_readable_time(run_duration),
        )
        self._orchestrator_run_id = None
        return None

    def submit_dynamic_pipeline(
        self,
        snapshot: "PipelineSnapshotResponse",
        stack: "Stack",
        environment: Dict[str, str],
        placeholder_run: Optional["PipelineRunResponse"] = None,
    ) -> Optional[SubmissionResult]:
        """Submits a dynamic pipeline to the orchestrator.

        Args:
            snapshot: The pipeline snapshot to submit.
            stack: The stack the pipeline will run on.
            environment: Environment variables to set in the orchestration
                environment.
            placeholder_run: An optional placeholder run.

        Returns:
            Optional submission result.
        """
        from zenml.execution.pipeline.dynamic.runner import (
            DynamicPipelineRunner,
        )

        self._orchestrator_run_id = str(uuid4())
        start_time = time.time()

        runner = DynamicPipelineRunner(
            snapshot=snapshot, run=placeholder_run, orchestrator=self
        )
        with temporary_environment(environment):
            runner.run_pipeline()

        run_duration = time.time() - start_time
        logger.info(
            "Pipeline run has finished in `%s`.",
            string_utils.get_human_readable_time(run_duration),
        )
        self._orchestrator_run_id = None
        return None

    def get_orchestrator_run_id(self) -> str:
        """Returns the active orchestrator run id.

        Raises:
            RuntimeError: If no run id exists. This happens when this method
                gets called while the orchestrator is not running a pipeline.

        Returns:
            The orchestrator run id.
        """
        if not self._orchestrator_run_id:
            raise RuntimeError("No run id set.")

        return self._orchestrator_run_id

    @property
    def supported_execution_modes(self) -> List[ExecutionMode]:
        """Returns the supported execution modes for this flavor.

        Returns:
            A tuple of supported execution modes.
        """
        return [
            ExecutionMode.FAIL_FAST,
            ExecutionMode.STOP_ON_FAILURE,
            ExecutionMode.CONTINUE_ON_FAILURE,
        ]


class LocalOrchestratorConfig(BaseOrchestratorConfig):
    """Local orchestrator config."""

    @property
    def is_local(self) -> bool:
        """Checks if this stack component is running locally.

        Returns:
            True if this config is for a local component, False otherwise.
        """
        return True

    @property
    def is_synchronous(self) -> bool:
        """Whether the orchestrator runs synchronous or not.

        Returns:
            Whether the orchestrator runs synchronous or not.
        """
        return True


class LocalOrchestratorFlavor(BaseOrchestratorFlavor):
    """Class for the `LocalOrchestratorFlavor`."""

    @property
    def name(self) -> str:
        """The flavor name.

        Returns:
            The flavor name.
        """
        return "local"

    @property
    def docs_url(self) -> Optional[str]:
        """A URL to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A URL to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A URL to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/orchestrator/local.png"

    @property
    def config_class(self) -> Type[BaseOrchestratorConfig]:
        """Config class for the base orchestrator flavor.

        Returns:
            The config class.
        """
        return LocalOrchestratorConfig

    @property
    def implementation_class(self) -> Type[LocalOrchestrator]:
        """Implementation class for this flavor.

        Returns:
            The implementation class for this flavor.
        """
        return LocalOrchestrator



#  Copyright (c) ZenML GmbH 2023. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Models representing tags."""

import random
from typing import TYPE_CHECKING, ClassVar, List, Optional, Type, TypeVar

from pydantic import Field, field_validator

from zenml.constants import STR_FIELD_MAX_LENGTH
from zenml.enums import ColorVariants, TaggableResourceTypes
from zenml.models.v2.base.base import BaseUpdate
from zenml.models.v2.base.scoped import (
    UserScopedFilter,
    UserScopedRequest,
    UserScopedResponse,
    UserScopedResponseBody,
    UserScopedResponseMetadata,
    UserScopedResponseResources,
)
from zenml.utils.uuid_utils import is_valid_uuid

if TYPE_CHECKING:
    from sqlalchemy.sql.elements import ColumnElement

    from zenml.zen_stores.schemas import BaseSchema

    AnySchema = TypeVar("AnySchema", bound="BaseSchema")

# ------------------ Request Model ------------------


class TagRequest(UserScopedRequest):
    """Request model for tags."""

    name: str = Field(
        description="The unique title of the tag.",
        max_length=STR_FIELD_MAX_LENGTH,
    )
    exclusive: bool = Field(
        description="The flag signifying whether the tag is an exclusive tag.",
        default=False,
    )
    color: ColorVariants = Field(
        description="The color variant assigned to the tag.",
        default_factory=lambda: random.choice(list(ColorVariants)),
    )

    @field_validator("name")
    @classmethod
    def validate_name_not_uuid(cls, value: str) -> str:
        """Validates that the tag name is not a UUID.

        Args:
            value: The tag name to validate.

        Returns:
            The validated tag name.

        Raises:
            ValueError: If the tag name can be converted
                to a UUID.
        """
        if is_valid_uuid(value):
            raise ValueError(
                "Tag names cannot be UUIDs or strings that "
                "can be converted to UUIDs."
            )
        return value


# ------------------ Update Model ------------------


class TagUpdate(BaseUpdate):
    """Update model for tags."""

    name: Optional[str] = None
    exclusive: Optional[bool] = None
    color: Optional[ColorVariants] = None

    @field_validator("name")
    @classmethod
    def validate_name_not_uuid(cls, value: Optional[str]) -> Optional[str]:
        """Validates that the tag name is not a UUID.

        Args:
            value: The tag name to validate.

        Returns:
            The validated tag name.

        Raises:
            ValueError: If the tag name can be converted to a UUID.
        """
        if value is not None and is_valid_uuid(value):
            raise ValueError(
                "Tag names cannot be UUIDs or strings that "
                "can be converted to UUIDs."
            )
        return value


# ------------------ Response Model ------------------


class TagResponseBody(UserScopedResponseBody):
    """Response body for tags."""

    color: ColorVariants = Field(
        description="The color variant assigned to the tag.",
        default_factory=lambda: random.choice(list(ColorVariants)),
    )
    exclusive: bool = Field(
        description="The flag signifying whether the tag is an exclusive tag."
    )


class TagResponseMetadata(UserScopedResponseMetadata):
    """Response metadata for tags."""

    tagged_count: int = Field(
        description="The count of resources tagged with this tag."
    )


class TagResponseResources(UserScopedResponseResources):
    """Class for all resource models associated with the tag entity."""


class TagResponse(
    UserScopedResponse[
        TagResponseBody, TagResponseMetadata, TagResponseResources
    ]
):
    """Response model for tags."""

    name: str = Field(
        description="The unique title of the tag.",
        max_length=STR_FIELD_MAX_LENGTH,
    )

    def get_hydrated_version(self) -> "TagResponse":
        """Get the hydrated version of this tag.

        Returns:
            an instance of the same entity with the metadata field attached.
        """
        from zenml.client import Client

        return Client().zen_store.get_tag(self.id)

    @property
    def color(self) -> ColorVariants:
        """The `color` property.

        Returns:
            the value of the property.
        """
        return self.get_body().color

    @property
    def exclusive(self) -> bool:
        """The `exclusive` property.

        Returns:
            the value of the property.
        """
        return self.get_body().exclusive

    @property
    def tagged_count(self) -> int:
        """The `tagged_count` property.

        Returns:
            the value of the property.
        """
        return self.get_metadata().tagged_count


# ------------------ Filter Model ------------------


class TagFilter(UserScopedFilter):
    """Model to enable advanced filtering of all tags."""

    FILTER_EXCLUDE_FIELDS: ClassVar[List[str]] = [
        *UserScopedFilter.FILTER_EXCLUDE_FIELDS,
        "resource_type",
    ]

    name: Optional[str] = Field(
        description="The unique title of the tag.", default=None
    )
    color: Optional[ColorVariants] = Field(
        description="The color variant assigned to the tag.", default=None
    )
    exclusive: Optional[bool] = Field(
        description="The flag signifying whether the tag is an exclusive tag.",
        default=None,
    )
    resource_type: Optional[TaggableResourceTypes] = Field(
        description="Filter tags associated with a specific resource type.",
        default=None,
    )

    def get_custom_filters(
        self, table: Type["AnySchema"]
    ) -> List["ColumnElement[bool]"]:
        """Get custom filters.

        Args:
            table: The query table.

        Returns:
            A list of custom filters.
        """
        custom_filters = super().get_custom_filters(table)

        from sqlmodel import exists, select

        from zenml.zen_stores.schemas import (
            TagResourceSchema,
            TagSchema,
        )

        if self.resource_type:
            # Filter for tags that have at least one association with the specified resource type
            resource_type_filter = exists(
                select(TagResourceSchema).where(
                    TagResourceSchema.tag_id == TagSchema.id,
                    TagResourceSchema.resource_type
                    == self.resource_type.value,
                )
            )
            custom_filters.append(resource_type_filter)

        return custom_filters



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Kubernetes pod settings."""

from typing import Any, Dict, List, Optional

from pydantic import field_validator

from zenml.config.base_settings import BaseSettings
from zenml.integrations.kubernetes import serialization_utils
from zenml.logger import get_logger

logger = get_logger(__name__)


_pod_settings_logged_warnings = []


def warn_if_invalid_model_data(data: Any, class_name: str) -> None:
    """Validates the data of a Kubernetes model.

    Args:
        data: The data to validate.
        class_name: Name of the class of the model.
    """
    if not isinstance(data, dict):
        return

    try:
        serialization_utils.deserialize_kubernetes_model(data, class_name)
    except KeyError as e:
        if str(e) not in _pod_settings_logged_warnings:
            _pod_settings_logged_warnings.append(str(e))
            logger.warning(
                "Invalid data for Kubernetes model class `%s`: %s. "
                "Hint: Kubernetes expects attribute names in CamelCase, not "
                "snake_case.",
                class_name,
                e,
            )


class KubernetesPodSettings(BaseSettings):
    """Kubernetes Pod settings.

    Attributes:
        node_selectors: Node selectors to apply to the pod.
        affinity: Affinity to apply to the pod.
        tolerations: Tolerations to apply to the pod.
        resources: Resource requests and limits for the pod.
        annotations: Annotations to apply to the pod metadata.
        volumes: Volumes to mount in the pod.
        volume_mounts: Volume mounts to apply to the pod containers.
        host_ipc: Whether to enable host IPC for the pod.
        scheduler_name: The name of the scheduler to use for the pod.
        image_pull_secrets: Image pull secrets to use for the pod.
        labels: Labels to apply to the pod.
        env: Environment variables to apply to the container.
        env_from: Environment variables to apply to the container.
        container_security_context: Security context settings to apply to all
            containers in the pod. This allows specifying container-level security
            attributes such as runAsUser, runAsNonRoot, allowPrivilegeEscalation,
            etc. Note: This is different from pod-level security context (which can
            be set via additional_pod_spec_args) as some admission policies require
            container-level security settings.
        additional_pod_spec_args: Additional arguments to pass to the pod. These
            will be applied to the pod spec.
    """

    node_selectors: Dict[str, str] = {}
    affinity: Dict[str, Any] = {}
    tolerations: List[Dict[str, Any]] = []
    resources: Dict[str, Dict[str, str]] = {}
    annotations: Dict[str, str] = {}
    volumes: List[Dict[str, Any]] = []
    volume_mounts: List[Dict[str, Any]] = []
    host_ipc: bool = False
    scheduler_name: Optional[str] = None
    image_pull_secrets: List[str] = []
    labels: Dict[str, str] = {}
    env: List[Dict[str, Any]] = []
    env_from: List[Dict[str, Any]] = []
    container_security_context: Dict[str, Any] = {}
    additional_pod_spec_args: Dict[str, Any] = {}

    @field_validator("volumes", mode="before")
    @classmethod
    def _convert_volumes(cls, value: Any) -> Any:
        """Converts Kubernetes volumes to dicts.

        Args:
            value: The volumes list.

        Returns:
            The converted volumes.
        """
        from kubernetes.client.models import V1Volume

        result = []
        for element in value:
            if isinstance(element, V1Volume):
                result.append(
                    serialization_utils.serialize_kubernetes_model(element)
                )
            else:
                warn_if_invalid_model_data(element, "V1Volume")
                result.append(element)

        return result

    @field_validator("volume_mounts", mode="before")
    @classmethod
    def _convert_volume_mounts(cls, value: Any) -> Any:
        """Converts Kubernetes volume mounts to dicts.

        Args:
            value: The volume mounts list.

        Returns:
            The converted volume mounts.
        """
        from kubernetes.client.models import V1VolumeMount

        result = []
        for element in value:
            if isinstance(element, V1VolumeMount):
                result.append(
                    serialization_utils.serialize_kubernetes_model(element)
                )
            else:
                warn_if_invalid_model_data(element, "V1VolumeMount")
                result.append(element)

        return result

    @field_validator("affinity", mode="before")
    @classmethod
    def _convert_affinity(cls, value: Any) -> Any:
        """Converts Kubernetes affinity to a dict.

        Args:
            value: The affinity value.

        Returns:
            The converted value.
        """
        from kubernetes.client.models import V1Affinity

        if isinstance(value, V1Affinity):
            return serialization_utils.serialize_kubernetes_model(value)
        else:
            warn_if_invalid_model_data(value, "V1Affinity")
            return value

    @field_validator("tolerations", mode="before")
    @classmethod
    def _convert_tolerations(cls, value: Any) -> Any:
        """Converts Kubernetes tolerations to dicts.

        Args:
            value: The tolerations list.

        Returns:
            The converted tolerations.
        """
        from kubernetes.client.models import V1Toleration

        result = []
        for element in value:
            if isinstance(element, V1Toleration):
                result.append(
                    serialization_utils.serialize_kubernetes_model(element)
                )
            else:
                warn_if_invalid_model_data(element, "V1Toleration")
                result.append(element)

        return result

    @field_validator("resources", mode="before")
    @classmethod
    def _convert_resources(cls, value: Any) -> Any:
        """Converts Kubernetes resource requirements to a dict.

        Args:
            value: The resource value.

        Returns:
            The converted value.
        """
        from kubernetes.client.models import V1ResourceRequirements

        if isinstance(value, V1ResourceRequirements):
            return serialization_utils.serialize_kubernetes_model(value)
        else:
            warn_if_invalid_model_data(value, "V1ResourceRequirements")
            return value

    @field_validator("env", mode="before")
    @classmethod
    def _convert_env(cls, value: Any) -> Any:
        """Converts Kubernetes EnvVar to a dict.

        Args:
            value: The env value.

        Returns:
            The converted value.
        """
        from kubernetes.client.models import V1EnvVar

        result = []
        for element in value:
            if isinstance(element, V1EnvVar):
                result.append(
                    serialization_utils.serialize_kubernetes_model(element)
                )
            else:
                warn_if_invalid_model_data(element, "V1EnvVar")
                result.append(element)

        return result

    @field_validator("env_from", mode="before")
    @classmethod
    def _convert_env_from(cls, value: Any) -> Any:
        """Converts Kubernetes EnvFromSource to a dict.

        Args:
            value: The env from value.

        Returns:
            The converted value.
        """
        from kubernetes.client.models import V1EnvFromSource

        result = []
        for element in value:
            if isinstance(element, V1EnvFromSource):
                result.append(
                    serialization_utils.serialize_kubernetes_model(element)
                )
            else:
                warn_if_invalid_model_data(element, "V1EnvFromSource")
                result.append(element)

        return result

    @field_validator("container_security_context", mode="before")
    @classmethod
    def _convert_container_security_context(cls, value: Any) -> Any:
        """Converts Kubernetes SecurityContext to a dict.

        Args:
            value: The container security context value.

        Returns:
            The converted value.
        """
        from kubernetes.client.models import V1SecurityContext

        if isinstance(value, V1SecurityContext):
            return serialization_utils.serialize_kubernetes_model(value)
        else:
            warn_if_invalid_model_data(value, "V1SecurityContext")
            return value



#  Copyright (c) ZenML GmbH 2025. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
#
# Parts of the `prepare_or_run_pipeline()` method of this file are
# inspired by the Kubernetes dag runner implementation of tfx
"""Kubernetes orchestrator constants."""

ENV_ZENML_KUBERNETES_RUN_ID = "ZENML_KUBERNETES_RUN_ID"
KUBERNETES_SECRET_TOKEN_KEY_NAME = "zenml_api_token"
KUBERNETES_CRON_JOB_METADATA_KEY = "cron_job_name"
# Annotation keys
ORCHESTRATOR_ANNOTATION_KEY = "zenml.io/orchestrator"
RUN_ID_ANNOTATION_KEY = "zenml.io/run-id"
ORCHESTRATOR_RUN_ID_ANNOTATION_KEY = "zenml.io/orchestrator-run-id"
STEP_NAME_ANNOTATION_KEY = "zenml.io/step-name"
STEP_OPERATOR_ANNOTATION_KEY = "zenml.io/step-operator"



#  Copyright (c) ZenML GmbH 2024. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""AWS Code Build image builder flavor."""

from typing import TYPE_CHECKING, Dict, Optional, Type

from zenml.image_builders import BaseImageBuilderConfig, BaseImageBuilderFlavor
from zenml.integrations.aws import (
    AWS_CONNECTOR_TYPE,
    AWS_IMAGE_BUILDER_FLAVOR,
    AWS_RESOURCE_TYPE,
)
from zenml.models import ServiceConnectorRequirements

if TYPE_CHECKING:
    from zenml.integrations.aws.image_builders import AWSImageBuilder


DEFAULT_CLOUDBUILD_IMAGE = "bentolor/docker-dind-awscli"
DEFAULT_CLOUDBUILD_COMPUTE_TYPE = "BUILD_GENERAL1_SMALL"


class AWSImageBuilderConfig(BaseImageBuilderConfig):
    """AWS Code Build image builder configuration.

    Attributes:
        code_build_project: The name of an existing AWS CodeBuild project to use
            to build the image. The CodeBuild project must exist in the AWS
            account and region inferred from the AWS service connector
            credentials or implicitly from the local AWS config.
        build_image: The Docker image to use for the AWS CodeBuild environment.
            The image must have Docker installed and be able to run Docker
            commands. The default image is bentolor/docker-dind-awscli.
            This can be customized to use a mirror, if needed, in case the
            Dockerhub image is not accessible or rate-limited.
        custom_env_vars: Custom environment variables to pass to the AWS
            CodeBuild build.
        compute_type: The compute type to use for the AWS CodeBuild build.
            The default is BUILD_GENERAL1_SMALL.
        implicit_container_registry_auth: Whether to use implicit authentication
            to authenticate the AWS Code Build build to the container registry
            when pushing container images. If set to False, the container
            registry credentials must be explicitly configured for the container
            registry stack component or the container registry stack component
            must be linked to a service connector.
            NOTE: When implicit_container_registry_auth is set to False, the
            container registry credentials will be passed to the AWS Code Build
            build as environment variables. This is not recommended for
            production use unless your service connector is configured to
            generate short-lived credentials.
    """

    code_build_project: str
    build_image: str = DEFAULT_CLOUDBUILD_IMAGE
    custom_env_vars: Optional[Dict[str, str]] = None
    compute_type: str = DEFAULT_CLOUDBUILD_COMPUTE_TYPE
    implicit_container_registry_auth: bool = True


class AWSImageBuilderFlavor(BaseImageBuilderFlavor):
    """AWS Code Build image builder flavor."""

    @property
    def name(self) -> str:
        """The flavor name.

        Returns:
            The name of the flavor.
        """
        return AWS_IMAGE_BUILDER_FLAVOR

    @property
    def display_name(self) -> str:
        """Display name of the flavor.

        Returns:
            The display name of the flavor.
        """
        return "AWS"

    @property
    def service_connector_requirements(
        self,
    ) -> Optional[ServiceConnectorRequirements]:
        """Service connector resource requirements for service connectors.

        Specifies resource requirements that are used to filter the available
        service connector types that are compatible with this flavor.

        Returns:
            Requirements for compatible service connectors, if a service
            connector is required for this flavor.
        """
        return ServiceConnectorRequirements(
            connector_type=AWS_CONNECTOR_TYPE,
            resource_type=AWS_RESOURCE_TYPE,
        )

    @property
    def docs_url(self) -> Optional[str]:
        """A url to point at docs explaining this flavor.

        Returns:
            A flavor docs url.
        """
        return self.generate_default_docs_url()

    @property
    def sdk_docs_url(self) -> Optional[str]:
        """A url to point at SDK docs explaining this flavor.

        Returns:
            A flavor SDK docs url.
        """
        return self.generate_default_sdk_docs_url()

    @property
    def logo_url(self) -> str:
        """A url to represent the flavor in the dashboard.

        Returns:
            The flavor logo.
        """
        return "https://public-flavor-logos.s3.eu-central-1.amazonaws.com/image_builder/aws.png"

    @property
    def config_class(self) -> Type[BaseImageBuilderConfig]:
        """The config class.

        Returns:
            The config class.
        """
        return AWSImageBuilderConfig

    @property
    def implementation_class(self) -> Type["AWSImageBuilder"]:
        """Implementation class.

        Returns:
            The implementation class.
        """
        from zenml.integrations.aws.image_builders import AWSImageBuilder

        return AWSImageBuilder



#  Copyright (c) ZenML GmbH 2021. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Validation functions for hooks."""

from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Tuple,
    Union,
)

from pydantic import ConfigDict, ValidationError

from zenml.config.source import Source
from zenml.exceptions import HookValidationException
from zenml.logger import get_logger
from zenml.utils import source_utils
from zenml.utils.pydantic_utils import validate_function_args

logger = get_logger(__name__)


if TYPE_CHECKING:
    from zenml.types import HookSpecification, InitHookSpecification


def _validate_hook_arguments(
    _func: Callable[..., Any],
    hook_kwargs: Dict[str, Any],
    exception_arg: Union[BaseException, bool] = False,
) -> Dict[str, Any]:
    """Validates hook arguments.

    Args:
        _func: The hook function to validate.
        hook_kwargs: The hook keyword arguments to validate.
        exception_arg: The exception argument to validate.

    Returns:
        The validated hook arguments.

    Raises:
        HookValidationException: If the hook arguments are not valid.
    """
    # Validate hook arguments
    try:
        hook_args: Tuple[Any, ...] = ()
        if isinstance(exception_arg, BaseException):
            hook_args = (exception_arg,)
        elif exception_arg is True:
            hook_args = (Exception(),)
        config = ConfigDict(arbitrary_types_allowed=len(hook_args) > 0)
        validated_kwargs = validate_function_args(
            _func, config, *hook_args, **hook_kwargs
        )
    except (ValidationError, TypeError) as e:
        exc_msg = (
            "Failed to validate hook arguments for {func}: {e}\n"
            "Please observe the following guidelines:\n"
            "- the success hook takes no arguments\n"
            "- the failure hook optionally takes a single `BaseException` "
            "typed argument\n"
            "- the init hook takes any number of JSON-safe arguments\n"
            "- the cleanup hook takes no arguments\n"
        )

        if not hook_args:
            raise HookValidationException(exc_msg.format(func=_func, e=e))

        # If we have an exception argument, we try again without it. This is
        # to account for the case where the hook function does not expect an
        # exception argument.
        hook_args = ()
        config = ConfigDict(arbitrary_types_allowed=False)
        try:
            validated_kwargs = validate_function_args(
                _func, config, *hook_args, **hook_kwargs
            )
        except (ValidationError, TypeError) as e:
            raise HookValidationException(exc_msg.format(func=_func, e=e))

    return validated_kwargs


def resolve_and_validate_hook(
    hook: Union["HookSpecification", "InitHookSpecification"],
    hook_kwargs: Optional[Dict[str, Any]] = None,
    allow_exception_arg: bool = False,
) -> Tuple[Source, Optional[Dict[str, Any]]]:
    """Resolves and validates a hook callback and its arguments.

    Args:
        hook: Hook function or source.
        hook_kwargs: The arguments to pass to the hook.
        allow_exception_arg: Whether to allow an implicit exception argument
            to be passed to the hook.

    Returns:
        Tuple of hook source and validated hook arguments converted to JSON-safe
        values.

    Raises:
        ValueError: If `hook_func` is not a valid callable.
    """
    # Resolve the hook function
    if isinstance(hook, (str, Source)):
        func = source_utils.load(hook)
    else:
        func = hook

    if not callable(func):
        raise ValueError(f"{func} is not a valid function.")

    # Validate hook arguments
    validated_kwargs = _validate_hook_arguments(
        func, hook_kwargs or {}, allow_exception_arg
    )

    return source_utils.resolve(func), validated_kwargs


def load_and_run_hook(
    hook_source: "Source",
    hook_parameters: Optional[Dict[str, Any]] = None,
    step_exception: Optional[BaseException] = None,
    raise_on_error: bool = False,
) -> Any:
    """Loads hook source and runs the hook.

    Args:
        hook_source: The source of the hook function.
        hook_parameters: The parameters of the hook function.
        step_exception: The exception of the original step.
        raise_on_error: Whether to raise an error if the hook fails.

    Returns:
        The return value of the hook function.

    Raises:
        HookValidationException: If hook validation fails.
        RuntimeError: If the hook fails and raise_on_error is True.
    """
    try:
        hook = source_utils.load(hook_source)
    except Exception as e:
        msg = f"Failed to load hook source '{hook_source}' with exception: {e}"
        if raise_on_error:
            raise RuntimeError(msg) from e
        else:
            logger.error(msg)
            return None
    try:
        validated_kwargs = _validate_hook_arguments(
            hook, hook_parameters or {}, step_exception or False
        )
    except HookValidationException as e:
        if raise_on_error:
            raise
        else:
            logger.error(e)
            return None

    try:
        logger.debug(f"Running hook {hook} with params: {validated_kwargs}")
        return hook(**validated_kwargs)
    except Exception as e:
        msg = (
            f"Failed to run hook '{hook_source}' with params: "
            f"{validated_kwargs} with exception: '{e}'"
        )
        if raise_on_error:
            raise RuntimeError(msg) from e
        else:
            logger.error(msg)
            return None



#  Copyright (c) ZenML GmbH 2022. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at:
#
#       https://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
#  or implied. See the License for the specific language governing
#  permissions and limitations under the License.
"""Utility functions for networking."""

import socket
from typing import List, Optional, Tuple, cast
from urllib.parse import urlparse

from zenml.environment import Environment
from zenml.logger import get_logger

logger = get_logger(__name__)

# default scanning port range for allocating ports
SCAN_PORT_RANGE = (8000, 65535)


def port_available(port: int, address: str = "127.0.0.1") -> bool:
    """Checks if a local port is available.

    Args:
        port: TCP port number
        address: IP address on the local machine

    Returns:
        True if the port is available, otherwise False
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if hasattr(socket, "SO_REUSEPORT"):
                s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
            else:
                # The SO_REUSEPORT socket option is not supported on Windows.
                # This if clause exists just for mypy to not complain about
                # missing code paths.
                pass
            s.bind((address, port))
    except socket.error as e:
        logger.debug("Port %d unavailable on %s: %s", port, address, e)
        return False

    return True


def find_available_port() -> int:
    """Finds a local random unoccupied TCP port.

    Returns:
        A random unoccupied TCP port.
    """
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        s.bind(("127.0.0.1", 0))
        _, port = s.getsockname()

    return cast(int, port)


def lookup_preferred_or_free_port(
    preferred_ports: List[int] = [],
    allocate_port_if_busy: bool = True,
    range: Tuple[int, int] = SCAN_PORT_RANGE,
    address: str = "127.0.0.1",
) -> int:
    """Find a preferred TCP port that is available or search for a free TCP port.

    If a list of preferred TCP port values is explicitly requested, they
    will be checked in order.

    Args:
        preferred_ports: A list of preferred TCP port values.
        allocate_port_if_busy: If True, allocate a free port if all the
            preferred ports are busy, otherwise an exception will be raised.
        range: The range of ports to search for a free port.
        address: The IP address for which to find a port.

    Returns:
        An available TCP port number

    Raises:
        IOError: if the preferred TCP port is busy and `allocate_port_if_busy`
            is disabled, or if no free TCP port could be otherwise allocated.
    """
    # If a port value is explicitly configured, attempt to use it first
    if preferred_ports:
        for port in preferred_ports:
            if port_available(port, address):
                return port
        if not allocate_port_if_busy:
            raise IOError(f"TCP port {preferred_ports} is not available.")

    available_port = scan_for_available_port(
        start=range[0], stop=range[1], address=address
    )
    if available_port:
        return available_port
    raise IOError(f"No free TCP ports found in range {range}")


def scan_for_available_port(
    start: int = SCAN_PORT_RANGE[0],
    stop: int = SCAN_PORT_RANGE[1],
    address: str = "127.0.0.1",
) -> Optional[int]:
    """Scan the local network for an available port in the given range.

    Args:
        start: the beginning of the port range value to scan
        stop: the (inclusive) end of the port range value to scan
        address: The IP address for which to find a port.

    Returns:
        The first available port in the given range, or None if no available
        port is found.
    """
    for port in range(start, stop + 1):
        if port_available(port, address):
            return port
    logger.debug(
        "No free TCP ports found in the range %d - %d",
        start,
        stop,
    )
    return None


def port_is_open(hostname: str, port: int) -> bool:
    """Check if a TCP port is open on a remote host.

    Args:
        hostname: hostname of the remote machine
        port: TCP port number

    Returns:
        True if the port is open, False otherwise
    """
    try:
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
            result = sock.connect_ex((hostname, port))
            return result == 0
    except socket.error as e:
        logger.debug(
            f"Error checking TCP port {port} on host {hostname}: {str(e)}"
        )
        return False


def replace_localhost_with_internal_hostname(url: str) -> str:
    """Replaces the localhost with an internal Docker or K3D hostname in a given URL.

    Localhost URLs that are directly accessible on the host machine are not
    accessible from within a Docker or K3D container running on that same
    machine, but there are special hostnames featured by both Docker
    (`host.docker.internal`) and K3D (`host.k3d.internal`) that can be used to
    access host services from within the containers.

    Use this method to attempt to replace `localhost` in a URL with one of these
    special hostnames, if they are available inside a container.

    Args:
        url: The URL to update.

    Returns:
        The updated URL.
    """
    if not Environment.in_container():
        return url

    parsed_url = urlparse(url)
    if parsed_url.hostname in ("localhost", "127.0.0.1"):
        for internal_hostname in (
            "host.docker.internal",
            "host.k3d.internal",
        ):
            try:
                socket.gethostbyname(internal_hostname)
                parsed_url = parsed_url._replace(
                    netloc=parsed_url.netloc.replace(
                        parsed_url.hostname,
                        internal_hostname,
                    )
                )
                logger.debug(
                    f"Replacing localhost with {internal_hostname} in URL: "
                    f"{url}"
                )
                return parsed_url.geturl()

            except socket.gaierror:
                continue

    return url


def replace_internal_hostname_with_localhost(hostname: str) -> str:
    """Replaces an internal Docker or K3D hostname with localhost.

    Localhost URLs that are directly accessible on the host machine are not
    accessible from within a Docker or K3D container running on that same
    machine, but there are special hostnames featured by both Docker
    (`host.docker.internal`) and K3D (`host.k3d.internal`) that can be used to
    access host services from within the containers.

    Use this method to replace one of these special hostnames with localhost
    if used outside a container or in a container where special hostnames are
    not available.

    Args:
        hostname: The hostname to replace.

    Returns:
        The original or replaced hostname.
    """
    if hostname not in ("host.docker.internal", "host.k3d.internal"):
        return hostname

    if Environment.in_container():
        # Try to resolve one of the special hostnames to see if it is available
        # inside the container and use that if it is.
        for internal_hostname in (
            "host.docker.internal",
            "host.k3d.internal",
        ):
            try:
                socket.gethostbyname(internal_hostname)
                if internal_hostname != hostname:
                    logger.debug(
                        f"Replacing internal hostname {hostname} with "
                        f"{internal_hostname}"
                    )
                return internal_hostname
            except socket.gaierror:
                continue

    logger.debug(f"Replacing internal hostname {hostname} with localhost.")

    return "127.0.0.1"


def get_or_create_ngrok_tunnel(ngrok_token: str, port: int) -> str:
    """Get or create an ngrok tunnel at the given port.

    Args:
        ngrok_token: The ngrok auth token.
        port: The port to tunnel.

    Returns:
        The public URL of the ngrok tunnel.

    Raises:
        ImportError: If the `pyngrok` package is not installed.
    """
    try:
        from pyngrok import ngrok as ngrok_client
    except ImportError:
        raise ImportError(
            "The `pyngrok` package is required to create ngrok tunnels. "
            "Please install it by running `pip install pyngrok`."
        )

    # Check if ngrok is already tunneling the port
    tunnels = ngrok_client.get_tunnels()
    for tunnel in tunnels:
        if tunnel.config and isinstance(tunnel.config, dict):
            tunnel_protocol = tunnel.config.get("proto")
            tunnel_port = tunnel.config.get("addr")
            if tunnel_protocol == "http" and tunnel_port == port:
                return str(tunnel.public_url)

    # Create new tunnel
    ngrok_client.set_auth_token(ngrok_token)
    return str(ngrok_client.connect(port).public_url)
